/* -*- Mode: c; tab-width: 4; -*-
 * -*- ex: ts=4 -*-
 */

/*
 * Copyright 2019 Seenaptec Ltd. All rights reserved.
 * SEENAPTEC PROPRIETARY/CONFIDENTIAL.
 *
 * $Id$
 */

/*
 * Asm.h    (D.Vinnyk)
 * Vipm/machine/arm/Asm.h
 */

/* ------------------------------------------------------------------------- */

#ifndef __VIPM_ARM_ASM__
# define __VIPM_ARM_ASM__

/******************************************************************************
 * %% BeginSection: includes
 */

/*
# ifdef __ARM_NEON__
#  include <arm_neon.h>
# endif
 */

# ifndef __VIPM_ARM_SDEFS__
#  include <Vipm/machine/arm/Sdefs.h>
# endif

# ifdef _A_ARM64
#  include <Vipm/machine/arm64/Asm.h>
# else /* _A_ARM32 */
#  include <Vipm/machine/arm32/Asm.h>
# endif


/******************************************************************************
 * %% BeginSection: function definitions
 */

/*
 * @{
 */

/*
 * [abstract arm asm]
 *
 *
 * Suffix [A,O]DB - [auto,base offset] decrement before
 * Suffix [A,O]DA - [auto,base offset] decrement after
 * Suffix [A,O]IA - [auto,base offset] increment after
 * Suffix [A,O]IB - [auto,base offset] increment before
 *
 *
 * Vector registers:
 * H - 2 byte register // ISA64
 * S - 4 byte register // ISA64
 * D - 8 byte register
 * Q - 16 byte register
 *
 *
 * UCF    - [update] conditional flag
 *
 * RR     - regular register: 1 (numerated register), [src] (named register)
 * CRR    - counter (regular register)
 * DRR    - destination [data] (regular register)
 * SRR    - source [data] (regular register)
 * PRR    - data pointer (regular register)
 * LHSRR  - [data] regular register
 * RHSRR  - [data] regular register
 * VRI    - vector register index (Q or D)
 * DVRI   - destination [data] vector register index (Q or D)
 * SVRI   - source [data] vector register index (Q or D)
 * LHSVRI - [data] vector register index (Q or D)
 * RHSVRI - [data] vector register index (Q or D)
 * Imm    - immediate value: #K
 */


/* ========================================================================= */

/*
 * Converting/Saturating operations
 *
 * signed  : (x) --> (-2**(sat-1) <= x <= 2**(sat-1) -1)
 * unsigned: (x) --> (0 <= x <= 2**sat -1)
 * @{
 */

/* ------------------------------------------------------------------------- */

    /*
     * rounding convert float to signed/unsigned integer
     */
# define _ARM_RND_FTOSI32(D,S)              _ARM_RND_FTOSIEX(32,32,D,S,N)
# define _ARM_RND_FTOUI32(D,S)              _ARM_RND_FTOUIEX(32,32,D,S,N)
# define _ARM_RND_FTOSI64(D,S)              _ARM_RND_FTOSIEX(64,32,D,S,N)
# define _ARM_RND_FTOUI64(D,S)              _ARM_RND_FTOUIEX(64,32,D,S,N)
# define _ARM_RND_FTOSI(D,S)                _ARM_RND_FTOSI32(D,S)
# define _ARM_RND_FTOUI(D,S)                _ARM_RND_FTOUI32(D,S)

    /*
     * rounding convert double to signed/unsigned integer
     */
# define _ARM_RND_F64TOSI32(D,S)            _ARM_RND_FTOSIEX(32,64,D,S,N)
# define _ARM_RND_F64TOUI32(D,S)            _ARM_RND_FTOUIEX(32,64,D,S,N)
# define _ARM_RND_F64TOSI64(D,S)            _ARM_RND_FTOSIEX(64,64,D,S,N)
# define _ARM_RND_F64TOUI64(D,S)            _ARM_RND_FTOUIEX(64,64,D,S,N)
# define _ARM_RND_F64TOSI(D,S)              _ARM_RND_F64TOSI32(D,S)
# define _ARM_RND_F64TOUI(D,S)              _ARM_RND_F64TOUI32(D,S)

/* @} */


/* ========================================================================= */

/*
 * Logical bitwise operations
 * @{
 *
 * not                                                       : NOT[I]     : DRR = ~SRR
 * not (updating the condition flags)                        : NOT[I]_UCF : DRR = ~SRR
 * and                                                       : AND[I]     : DRR = SRR & N
 * and (updating the condition flags)                        : AND[I]_UCF : DRR = SRR & N
 * and complement (and * not)                                : BIC[I]     : DRR = SRR & ~N
 * and complement (and * not) (updating the condition flags) : BIC[I]_UCF : DRR = SRR & ~N
 * or                                                        : OR[I]      : DRR = SRR | N
 * or (updating the condition flags)                         : OR[I]_UCF  : DRR = SRR | N
 * xor                                                       : XOR[I]     : DRR = SRR ^ N
 * xor (updating the condition flags)                        : XOR[I]_UCF : DRR = SRR ^ N
 * set all bits                                              : BITSET     :
 * clear all bits                                            : BITCLR     :
 */

# define _ARM_BITSET(DRR)               _ARM_NOTI(DRR,0)
# define _ARM_BITSET_IF(_Cond,DRR)      _ARM_NOTI_IF(_Cond,DRR,0)

# define _ARM_BITCLR(DRR)               _ARM_BIC(DRR,DRR,DRR)
# define _ARM_BITCLR_IF(_Cond,DRR)      _ARM_BIC_IF(_Cond,DRR,DRR,DRR)

/* @} */


/* ========================================================================= */

/*
 * Memory access
 * @{
 */

    /*
     * Load
     */

# define _ARM_LOAD_PTR(DRR,PRR)                 _ARM_LOAD(_ARM_WIDTHOF_PTR, DRR,PRR)
# define _ARM_LOADO_PTR(DRR,PRR,ORR)            _ARM_LOADO(_ARM_WIDTHOF_PTR, DRR,PRR,ORR)
# define _ARM_LOADOI_PTR(DRR,PRR,Off)           _ARM_LOADOI(_ARM_WIDTHOF_PTR, DRR,PRR,_ARM_PTR_OFFSET(Off))

# define _ARM_LOAD_PTR_IF(_Cond, DRR,PRR)       _ARM_LOAD_IF(_ARM_WIDTHOF_PTR,_Cond, DRR,PRR)
# define _ARM_LOADO_PTR_IF(_Cond, DRR,PRR,ORR)  _ARM_LOADO_IF(_ARM_WIDTHOF_PTR,_Cond, DRR,PRR,ORR)
# define _ARM_LOADOI_PTR_IF(_Cond, DRR,PRR,Off) _ARM_LOADOI_IF(_ARM_WIDTHOF_PTR,_Cond, DRR,PRR,_ARM_PTR_OFFSET(Off))


/* slow operation (?) */
# define _ARM_LOAD_AIA(W, DRR, PRR)     _ARM_LOADI_IA(W, DRR, PRR, _ARM_W2SZ(W))
# define _ARM_LOAD_AIB(W, DRR, PRR)     _ARM_LOADI_IB(W, DRR, PRR, _ARM_W2SZ(W))
# define _ARM_LOAD_ADA(W, DRR, PRR)     _ARM_LOADI_DA(W, DRR, PRR, _ARM_W2SZ(W))
# define _ARM_LOAD_ADB(W, DRR, PRR)     _ARM_LOADI_DB(W, DRR, PRR, _ARM_W2SZ(W))

/* ------------------------------------------------------------------------- */

    /*
     * Store
     */

# define _ARM_STORE_PTR(DRR,PRR)                _ARM_STORE(_ARM_WIDTHOF_PTR, DRR,PRR)
# define _ARM_STOREO_PTR(DRR,PRR,ORR)           _ARM_STOREO(_ARM_WIDTHOF_PTR, DRR,PRR,ORR)
# define _ARM_STOREOI_PTR(DRR,PRR,Off)          _ARM_STOREOI(_ARM_WIDTHOF_PTR, DRR,PRR,_ARM_PTR_OFFSET(Off))

# define _ARM_STORE_PTR_IF(_Cond, DRR,PRR)      _ARM_STORE_IF(_ARM_WIDTHOF_PTR,_Cond, DRR,PRR)
# define _ARM_STOREO_PTR_IF(_Cond, DRR,PRR,ORR) _ARM_STOREO_IF(_ARM_WIDTHOF_PTR,_Cond, DRR,PRR,ORR)
# define _ARM_STOREOI_PTR_IF(_Cond, DRR,PRR,Off)_ARM_STOREOI_IF(_ARM_WIDTHOF_PTR,_Cond, DRR,PRR,_ARM_PTR_OFFSET(Off))


/* slow operation (?) */
# define _ARM_STORE_AIA(W, DRR, PRR)    _ARM_STOREI_IA(W, DRR, PRR, _ARM_W2SZ(W))
# define _ARM_STORE_AIB(W, DRR, PRR)    _ARM_STOREI_IB(W, DRR, PRR, _ARM_W2SZ(W))
# define _ARM_STORE_ADA(W, DRR, PRR)    _ARM_STOREI_DA(W, DRR, PRR, _ARM_W2SZ(W))
# define _ARM_STORE_ADB(W, DRR, PRR)    _ARM_STOREI_DB(W, DRR, PRR, _ARM_W2SZ(W))

/* ------------------------------------------------------------------------- */

    /*
     * Prefetch Memory
     */

# define _ARM_PRELOADI_READ(PRR,Offset,_Size)           _ARM_PREFETCHI(RD,L1,KEEP,PRR,Offset,_Size)
# define _ARM_PRELOADI_WRITE(PRR,Offset,_Size)          _ARM_PREFETCHI(WR,L1,KEEP,PRR,Offset,_Size)
# define _ARM_APRELOADI_READ(PRR,Offset,_Size)          _ARM_APREFETCHI(RD,L1,KEEP,PRR,Offset,_Size)
# define _ARM_APRELOADI_WRITE(PRR,Offset,_Size)         _ARM_APREFETCHI(WR,L1,KEEP,PRR,Offset,_Size)
# define _ARM_PRELOAD_READ(PRR,ORR,_Size)               _ARM_PREFETCH(RD,L1,KEEP,PRR,ORR,_Size)
# define _ARM_PRELOAD_WRITE(PRR,ORR,_Size)              _ARM_PREFETCH(WR,L1,KEEP,PRR,ORR,_Size)
# define _ARM_PRELOADEX_READ(PRR,ORR,Ext,_Size)         _ARM_PREFETCHEX(RD,L1,KEEP,PRR,ORR,Ext,_Size)
# define _ARM_PRELOADEX_WRITE(PRR,ORR,Ext,_Size)        _ARM_PREFETCHEX(WR,L1,KEEP,PRR,ORR,Ext,_Size)

# define _ARM_PRELOADI_READ_CACHELINE(PRR,Offset)       _ARM_PREFETCHI(RD,L1,KEEP,PRR,Offset,__VIPM_ARM_CACHE_LINE_SIZE)
# define _ARM_PRELOADI_WRITE_CACHELINE(PRR,Offset)      _ARM_PREFETCHI(WR,L1,KEEP,PRR,Offset,__VIPM_ARM_CACHE_LINE_SIZE)
# define _ARM_APRELOADI_READ_CACHELINE(PRR,Offset)      _ARM_APREFETCHI(RD,L1,KEEP,PRR,Offset,__VIPM_ARM_CACHE_LINE_SIZE)
# define _ARM_APRELOADI_WRITE_CACHELINE(PRR,Offset)     _ARM_APREFETCHI(WR,L1,KEEP,PRR,Offset,__VIPM_ARM_CACHE_LINE_SIZE)
# define _ARM_PRELOAD_READ_CACHELINE(PRR,ORR)           _ARM_PREFETCH(RD,L1,KEEP,PRR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)
# define _ARM_PRELOAD_WRITE_CACHELINE(PRR,ORR)          _ARM_PREFETCH(WR,L1,KEEP,PRR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)
# define _ARM_PRELOADEX_READ_CACHELINE(PRR,ORR,Ext)     _ARM_PREFETCHEX(RD,L1,KEEP,PRR,ORR,Ext,__VIPM_ARM_CACHE_LINE_SIZE)
# define _ARM_PRELOADEX_WRITE_CACHELINE(PRR,ORR,Ext)    _ARM_PREFETCHEX(WR,L1,KEEP,PRR,ORR,Ext,__VIPM_ARM_CACHE_LINE_SIZE)

/*
 * _Type:   {RD,WR,RW}
 * _Target: {L1:default(?),L2,L3}
 * _Policy: {KEEP:default(?),STRM}
 */
# define _ARM_PREFETCHI(_Type,_Target,_Policy,PRR,Offset,_Size)     VIPM_GLUE(_ARM_PREFETCHI_SZ,_Size)(_Type,_Target,_Policy,PRR,Offset)
# define _ARM_APREFETCHI(_Type,_Target,_Policy,PRR,Offset,_Size)    VIPM_GLUE(_ARM_APREFETCHI_SZ,_Size)(_Type,_Target,_Policy,PRR,Offset)
# define _ARM_PREFETCH(_Type,_Target,_Policy,PRR,ORR,_Size)         VIPM_GLUE(_ARM_PREFETCH_SZ,_Size)(_Type,_Target,_Policy,PRR,ORR)
# define _ARM_PREFETCHEX(_Type,_Target,_Policy,PRR,ORR,Ext,_Size)   VIPM_GLUE(_ARM_PREFETCHEX_SZ,_Size)(_Type,_Target,_Policy,PRR,ORR,Ext)


# if (__VIPM_ARM_CACHE_LINE_SIZE == 128)

#  define _ARM_PREFETCHI_SZ32(_Type,_Target,_Policy,PRR,Offset)     _ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_PREFETCHI_SZ64(_Type,_Target,_Policy,PRR,Offset)     _ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_PREFETCHI_SZ96(_Type,_Target,_Policy,PRR,Offset)     _ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_PREFETCHI_SZ128(_Type,_Target,_Policy,PRR,Offset)    _ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)

#  define _ARM_APREFETCHI_SZ32(_Type,_Target,_Policy,PRR,Offset)    _ARM_APREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_APREFETCHI_SZ64(_Type,_Target,_Policy,PRR,Offset)    _ARM_APREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_APREFETCHI_SZ96(_Type,_Target,_Policy,PRR,Offset)    _ARM_APREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_APREFETCHI_SZ128(_Type,_Target,_Policy,PRR,Offset)   _ARM_APREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)

#  define _ARM_PREFETCH_SZ32(_Type,_Target,_Policy,PRR,ORR)         _ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)
#  define _ARM_PREFETCH_SZ64(_Type,_Target,_Policy,PRR,ORR)         _ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)
#  define _ARM_PREFETCH_SZ96(_Type,_Target,_Policy,PRR,ORR)         _ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)
#  define _ARM_PREFETCH_SZ128(_Type,_Target,_Policy,PRR,ORR)        _ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)

#  define _ARM_PREFETCHEX_SZ32(_Type,_Target,_Policy,PRR,ORR,Ext)   _ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)
#  define _ARM_PREFETCHEX_SZ64(_Type,_Target,_Policy,PRR,ORR,Ext)   _ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)
#  define _ARM_PREFETCHEX_SZ96(_Type,_Target,_Policy,PRR,ORR,Ext)   _ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)
#  define _ARM_PREFETCHEX_SZ128(_Type,_Target,_Policy,PRR,ORR,Ext)  _ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)

# else
# if (__VIPM_ARM_CACHE_LINE_SIZE == 64)

#  define _ARM_PREFETCHI_SZ32(_Type,_Target,_Policy,PRR,Offset)     _ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_PREFETCHI_SZ64(_Type,_Target,_Policy,PRR,Offset)     _ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_PREFETCHI_SZ96(_Type,_Target,_Policy,PRR,Offset) \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)  \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,_ARM_PRFM_IADD64(Offset))
#  define _ARM_PREFETCHI_SZ128(_Type,_Target,_Policy,PRR,Offset) \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)   \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,_ARM_PRFM_IADD64(Offset))

#  define _ARM_APREFETCHI_SZ32(_Type,_Target,_Policy,PRR,Offset)    _ARM_APREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_APREFETCHI_SZ64(_Type,_Target,_Policy,PRR,Offset)    _ARM_APREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_APREFETCHI_SZ96(_Type,_Target,_Policy,PRR,Offset)    _ARM_PREFETCHI_SZ96(_Type,_Target,_Policy,PRR,_ARM_PRFM_OFFSET(Offset))
#  define _ARM_APREFETCHI_SZ128(_Type,_Target,_Policy,PRR,Offset)   _ARM_PREFETCHI_SZ128(_Type,_Target,_Policy,PRR,_ARM_PRFM_OFFSET(Offset))

#  define _ARM_PREFETCH_SZ32(_Type,_Target,_Policy,PRR,ORR)         _ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)
#  define _ARM_PREFETCH_SZ64(_Type,_Target,_Policy,PRR,ORR)         _ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)
#  define _ARM_PREFETCH_SZ96(_Type,_Target,_Policy,PRR,ORR) \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)  \
	_ARM_ADDI(ORR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)           \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)  \
	_ARM_SUBI(ORR,ORR,64)
#  define _ARM_PREFETCH_SZ128(_Type,_Target,_Policy,PRR,ORR) \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)   \
	_ARM_ADDI(ORR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)            \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)   \
	_ARM_SUBI(ORR,ORR,64)

#  define _ARM_PREFETCHEX_SZ32(_Type,_Target,_Policy,PRR,ORR,Ext)   _ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)
#  define _ARM_PREFETCHEX_SZ64(_Type,_Target,_Policy,PRR,ORR,Ext)   _ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)
#  define _ARM_PREFETCHEX_SZ96(_Type,_Target,_Policy,PRR,ORR,Ext) \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)  \
	_ARM_ADDI(PRR,PRR,__VIPM_ARM_CACHE_LINE_SIZE)                 \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)  \
	_ARM_SUBI(PRR,PRR,64)
#  define _ARM_PREFETCHEX_SZ128(_Type,_Target,_Policy,PRR,ORR,Ext) \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)   \
	_ARM_ADDI(PRR,PRR,__VIPM_ARM_CACHE_LINE_SIZE)                  \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)   \
	_ARM_SUBI(PRR,PRR,64)

# else
# if (__VIPM_ARM_CACHE_LINE_SIZE == 32)

#  define _ARM_PREFETCHI_SZ32(_Type,_Target,_Policy,PRR,Offset)     _ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_PREFETCHI_SZ64(_Type,_Target,_Policy,PRR,Offset) \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)  \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,_ARM_PRFM_IADD32(Offset))
#  define _ARM_PREFETCHI_SZ96(_Type,_Target,_Policy,PRR,Offset)                  \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)                   \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,_ARM_PRFM_IADD32(Offset)) \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,_ARM_PRFM_IADD64(Offset))
#  define _ARM_PREFETCHI_SZ128(_Type,_Target,_Policy,PRR,Offset)                 \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)                   \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,_ARM_PRFM_IADD32(Offset)) \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,_ARM_PRFM_IADD64(Offset)) \
	_ARM_PREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,_ARM_PRFM_IADD96(Offset))

#  define _ARM_APREFETCHI_SZ32(_Type,_Target,_Policy,PRR,Offset)    _ARM_APREFETCHI_CACHELINE(_Type,_Target,_Policy,PRR,Offset)
#  define _ARM_APREFETCHI_SZ64(_Type,_Target,_Policy,PRR,Offset)    _ARM_PREFETCHI_SZ64(_Type,_Target,_Policy,PRR,_ARM_PRFM_OFFSET(Offset))
#  define _ARM_APREFETCHI_SZ96(_Type,_Target,_Policy,PRR,Offset)    _ARM_PREFETCHI_SZ96(_Type,_Target,_Policy,PRR,_ARM_PRFM_OFFSET(Offset))
#  define _ARM_APREFETCHI_SZ128(_Type,_Target,_Policy,PRR,Offset)   _ARM_PREFETCHI_SZ128(_Type,_Target,_Policy,PRR,_ARM_PRFM_OFFSET(Offset))

#  define _ARM_PREFETCH_SZ32(_Type,_Target,_Policy,PRR,ORR)         _ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)
#  define _ARM_PREFETCH_SZ64(_Type,_Target,_Policy,PRR,ORR) \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)  \
	_ARM_ADDI(ORR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)           \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)  \
	_ARM_SUBI(ORR,ORR,32)
#  define _ARM_PREFETCH_SZ96(_Type,_Target,_Policy,PRR,ORR) \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)  \
	_ARM_ADDI(ORR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)           \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)  \
	_ARM_ADDI(ORR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)           \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)  \
	_ARM_SUBI(ORR,ORR,64)
#  define _ARM_PREFETCH_SZ128(_Type,_Target,_Policy,PRR,ORR) \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)   \
	_ARM_ADDI(ORR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)            \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)   \
	_ARM_ADDI(ORR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)            \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)   \
	_ARM_ADDI(ORR,ORR,__VIPM_ARM_CACHE_LINE_SIZE)            \
	_ARM_PREFETCH_CACHELINE(_Type,_Target,_Policy,PRR,ORR)   \
	_ARM_SUBI(ORR,ORR,96)

#  define _ARM_PREFETCHEX_SZ32(_Type,_Target,_Policy,PRR,ORR,Ext)   _ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)
#  define _ARM_PREFETCHEX_SZ64(_Type,_Target,_Policy,PRR,ORR,Ext) \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)  \
	_ARM_ADDI(PRR,PRR,__VIPM_ARM_CACHE_LINE_SIZE)                 \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)  \
	_ARM_SUBI(PRR,PRR,32)
#  define _ARM_PREFETCHEX_SZ96(_Type,_Target,_Policy,PRR,ORR,Ext) \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)  \
	_ARM_ADDI(PRR,PRR,__VIPM_ARM_CACHE_LINE_SIZE)                 \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)  \
	_ARM_ADDI(PRR,PRR,__VIPM_ARM_CACHE_LINE_SIZE)                 \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)  \
	_ARM_SUBI(PRR,PRR,64)
#  define _ARM_PREFETCHEX_SZ128(_Type,_Target,_Policy,PRR,ORR,Ext) \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)   \
	_ARM_ADDI(PRR,PRR,__VIPM_ARM_CACHE_LINE_SIZE)                  \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)   \
	_ARM_ADDI(PRR,PRR,__VIPM_ARM_CACHE_LINE_SIZE)                  \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)   \
	_ARM_ADDI(PRR,PRR,__VIPM_ARM_CACHE_LINE_SIZE)                  \
	_ARM_PREFETCHEX_CACHELINE(_Type,_Target,_Policy,PRR,ORR,Ext)   \
	_ARM_SUBI(PRR,PRR,96)

# endif /* (__VIPM_ARM_CACHE_LINE_SIZE == 32) */
# endif /* (__VIPM_ARM_CACHE_LINE_SIZE == 64) */
# endif /* (__VIPM_ARM_CACHE_LINE_SIZE == 128) */

/* @} */


/* ========================================================================= */

/*
 * Flow control
 * @{
 */

# define _ARM_XGOTO_IF(_Cond,_Stmt,Label)   _Stmt _ARM_GOTO_IF(_Cond,Label)

    /*
	 *  {
     *  } {* block-name: Label *}
     */
# define _ARM_ENDIF(Label)                  VIPM_STR_C(Label) ": \n\t"

/* ------------------------------------------------------------------------- */

    /*
     * do { ... } while (CRR -= Imm)
     * do { ... } while (CRR -= RR)
     * do { ... } while (0 <= (CRR -= Imm))
     * do { ... } while (0 <= (CRR -= RR))
     * do { ... } while ((CRR ?? Comparand) ? (CRR -= Value, 1) : 0)
     */

# define _ARM_DO(Label)                         VIPM_STR_C(Label) ": \n\t"

# define _ARM_WHILE(_Cond,Label)                _ARM_GOTO_IF(_Cond,Label)
# define _ARM_WHILE_TRUE(Label)                 _ARM_GOTO_IMM(Label)

# define _ARM_WHILEI_GT0_DB(CRR, Imm, Label)    _ARM_WHILEI_DB(GT, CRR, Imm, Label)
# define _ARM_WHILEI_GE0_DB(CRR, Imm, Label)    _ARM_WHILEI_DB(GE, CRR, Imm, Label)
# define _ARM_WHILEI_NE0_DB(CRR, Imm, Label)    _ARM_WHILEI_DB(NE, CRR, Imm, Label)
# define _ARM_WHILE_GT0_DB(CRR, RR, Label)      _ARM_WHILE_DB(GT, CRR, RR, Label)
# define _ARM_WHILE_GE0_DB(CRR, RR, Label)      _ARM_WHILE_DB(GE, CRR, RR, Label)
# define _ARM_WHILE_NE0_DB(CRR, RR, Label)      _ARM_WHILE_DB(NE, CRR, RR, Label)

# define _ARM_WHILEI_DB(_Cond, CRR,Imm, Label)  _ARM_SUBI_UCF(CRR,CRR,Imm) _ARM_GOTO_IF(_Cond,Label)
# define _ARM_WHILE_DB(_Cond, CRR,RR, Label)    _ARM_SUB_UCF(CRR,CRR,RR)   _ARM_GOTO_IF(_Cond,Label)

# define _ARM_WHILEI_DA(_Cond, CRR,Imm, Label)  _ARM_SUBI_UCF(CRR,CRR,Imm) _ARM_GOTO_IF(_Cond,Label) _ARM_ADDI(CRR,CRR,Imm)
# define _ARM_WHILE_DA(_Cond, CRR,RR, Label)    _ARM_SUB_UCF(CRR,CRR,RR)   _ARM_GOTO_IF(_Cond,Label) _ARM_ADD(CRR,CRR,RR)


# define _ARM_XWHILEI12_DA(_Cond, CRR,NI,Imm, Label) \
    _ARM_CMPI(CRR,NI)                                \
    _ARM_SUBI_IF(_Cond, CRR,CRR,Imm)                 \
    _ARM_GOTO_IF(_Cond,Label)

# define _ARM_XWHILEI10_DA(_Cond, CRR,NI,RR, Label) \
    _ARM_CMPI(CRR,NI)                               \
    _ARM_SUB_IF(_Cond, CRR,CRR,RR)                  \
    _ARM_GOTO_IF(_Cond,Label)

# define _ARM_XWHILEI02_DA(_Cond, CRR,N,Imm, Label) \
    _ARM_CMP(CRR,N)                                 \
    _ARM_SUBI_IF(_Cond, CRR,CRR,Imm)                \
    _ARM_GOTO_IF(_Cond,Label)

# define _ARM_XWHILE_DA(_Cond, CRR,N,RR, Label) \
    _ARM_CMP(CRR,N)                             \
    _ARM_SUB_IF(_Cond, CRR,CRR,RR)              \
    _ARM_GOTO_IF(_Cond,Label)

/* ------------------------------------------------------------------------- */

    /*
     * ADD
     */

# define _ARM_IF_ADDI(_Cond, DRR,SRR,Imm,Label)  _ARM_ADDI_UCF(DRR,SRR,Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)
# define _ARM_IF_ADD(_Cond, DRR,SRR1,SRR2,Label) _ARM_ADD_UCF(DRR,SRR1,SRR2) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)

# define _ARM_IF_ADDI_NE0(DRR,SRR,Imm, Label)   _ARM_IF_ADDI(NE, DRR, SRR, Imm, Label)
# define _ARM_IF_ADDI_GE0(DRR,SRR,Imm, Label)   _ARM_IF_ADDI(GE, DRR, SRR, Imm, Label)
# define _ARM_IF_ADDI_GT0(DRR,SRR,Imm, Label)   _ARM_IF_ADDI(GT, DRR, SRR, Imm, Label)

# define _ARM_IF_ADD_NE0(DRR,SRR1,SRR2, Label)  _ARM_IF_ADD(NE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_ADD_GE0(DRR,SRR1,SRR2, Label)  _ARM_IF_ADD(GE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_ADD_GT0(DRR,SRR1,SRR2, Label)  _ARM_IF_ADD(GT, DRR, SRR1, SRR2, Label)

/* ------------------------------------------------------------------------- */

    /*
     * SUB
     */

# define _ARM_IF_SUBI(_Cond, DRR,SRR,Imm,Label)  _ARM_SUBI_UCF(DRR,SRR,Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)
# define _ARM_IF_SUB(_Cond, DRR,SRR1,SRR2,Label) _ARM_SUB_UCF(DRR,SRR1,SRR2) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)

# define _ARM_IF_SUBI_NE0(DRR,SRR,Imm, Label)   _ARM_IF_SUBI(NE, DRR, SRR, Imm, Label)
# define _ARM_IF_SUBI_GE0(DRR,SRR,Imm, Label)   _ARM_IF_SUBI(GE, DRR, SRR, Imm, Label)
# define _ARM_IF_SUBI_GT0(DRR,SRR,Imm, Label)   _ARM_IF_SUBI(GT, DRR, SRR, Imm, Label)

# define _ARM_IF_SUB_NE0(DRR,SRR1,SRR2, Label)  _ARM_IF_SUB(NE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_SUB_GE0(DRR,SRR1,SRR2, Label)  _ARM_IF_SUB(GE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_SUB_GT0(DRR,SRR1,SRR2, Label)  _ARM_IF_SUB(GT, DRR, SRR1, SRR2, Label)

/* ------------------------------------------------------------------------- */

    /*
     * RSUB
     */

# define _ARM_IF_RSUBI(_Cond, DRR,SRR,Imm,Label)  _ARM_RSUBI_UCF(DRR,SRR,Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)
# define _ARM_IF_RSUB(_Cond, DRR,SRR1,SRR2,Label) _ARM_RSUB_UCF(DRR,SRR1,SRR2) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)

# define _ARM_IF_RSUBI_NE0(DRR,SRR,Imm, Label)  _ARM_IF_RSUBI(NE, DRR, SRR, Imm, Label)
# define _ARM_IF_RSUBI_GE0(DRR,SRR,Imm, Label)  _ARM_IF_RSUBI(GE, DRR, SRR, Imm, Label)
# define _ARM_IF_RSUBI_GT0(DRR,SRR,Imm, Label)  _ARM_IF_RSUBI(GT, DRR, SRR, Imm, Label)

# define _ARM_IF_RSUB_NE0(DRR,SRR1,SRR2, Label) _ARM_IF_RSUB(NE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_RSUB_GE0(DRR,SRR1,SRR2, Label) _ARM_IF_RSUB(GE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_RSUB_GT0(DRR,SRR1,SRR2, Label) _ARM_IF_RSUB(GT, DRR, SRR1, SRR2, Label)

/* ------------------------------------------------------------------------- */

    /*
     *  if {* block-name: Label *} ((DRR = (SRR & Imm)) != 0) { ...
	 *  }
     */

# define _ARM_IF_ANDI(_Cond, DRR,SRR,Imm,Label)   _ARM_ANDI_UCF(DRR,SRR,Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)
# define _ARM_IF_AND(_Cond, DRR,SRR1,SRR2,Label)  _ARM_AND_UCF(DRR,SRR1,SRR2) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)

# define _ARM_IF_ANDI_NE0(DRR,SRR,Imm, Label)   _ARM_IF_ANDI(NE, DRR, SRR, Imm, Label)
# define _ARM_IF_ANDI_GE0(DRR,SRR,Imm, Label)   _ARM_IF_ANDI(GE, DRR, SRR, Imm, Label)
# define _ARM_IF_ANDI_GT0(DRR,SRR,Imm, Label)   _ARM_IF_ANDI(GT, DRR, SRR, Imm, Label)

# define _ARM_IF_AND_NE0(DRR,SRR1,SRR2, Label)  _ARM_IF_AND(NE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_AND_GE0(DRR,SRR1,SRR2, Label)  _ARM_IF_AND(GE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_AND_GT0(DRR,SRR1,SRR2, Label)  _ARM_IF_AND(GT, DRR, SRR1, SRR2, Label)

/* ------------------------------------------------------------------------- */

    /*
     *  if {* block-name: Label *} ((DRR = (SRR ^ Imm)) != 0) { ...
	 *  }
     */

# define _ARM_IF_XORI(_Cond, DRR,SRR,Imm, Label)   ARM_XORI_UCF(DRR,SRR,Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)
# define _ARM_IF_XOR(_Cond, DRR,SRR1,SRR2, Label) _ARM_XOR_UCF(DRR,SRR1,SRR2) ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)

# define _ARM_IF_XORI_NE0(DRR,SRR,Imm, Label)     _ARM_IF_XORI(NE, DRR, SRR, Imm, Label)
# define _ARM_IF_XORI_GE0(DRR,SRR,Imm, Label)     _ARM_IF_XORI(GE, DRR, SRR, Imm, Label)
# define _ARM_IF_XORI_GT0(DRR,SRR,Imm, Label)     _ARM_IF_XORI(GT, DRR, SRR, Imm, Label)

# define _ARM_IF_XOR_NE0(DRR,SRR1,SRR2, Label)    _ARM_IF_XOR(NE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_XOR_GE0(DRR,SRR1,SRR2, Label)    _ARM_IF_XOR(GE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_XOR_GT0(DRR,SRR1,SRR2, Label)    _ARM_IF_XOR(GT, DRR, SRR1, SRR2, Label)

/* ------------------------------------------------------------------------- */

    /*
     *  if {* block-name: Label *} ((DRR = (SRR >> Imm)) != 0) { ...
	 *  }
     */

# define _ARM_IF_RSHIFTI(_Cond, DRR,SRR,Imm,Label)  _ARM_RSHIFTI_UCF(DRR,SRR,Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)
# define _ARM_IF_RSHIFT(_Cond, DRR,SRR1,SRR2,Label) _ARM_RSHIFT_UCF(DRR,SRR1,SRR2) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)

# define _ARM_IF_RSHIFTI_NE0(DRR, SRR, Imm, Label)  _ARM_IF_RSHIFTI(NE, DRR, SRR, Imm, Label)
# define _ARM_IF_RSHIFTI_GE0(DRR, SRR, Imm, Label)  _ARM_IF_RSHIFTI(GE, DRR, SRR, Imm, Label)
# define _ARM_IF_RSHIFTI_GT0(DRR, SRR, Imm, Label)  _ARM_IF_RSHIFTI(GT, DRR, SRR, Imm, Label)

# define _ARM_IF_RSHIFT_NE0(DRR, SRR1, SRR2, Label) _ARM_IF_RSHIFT(NE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_RSHIFT_GE0(DRR, SRR1, SRR2, Label) _ARM_IF_RSHIFT(GE, DRR, SRR1, SRR2, Label)
# define _ARM_IF_RSHIFT_GT0(DRR, SRR1, SRR2, Label) _ARM_IF_RSHIFT(GT, DRR, SRR1, SRR2, Label)


# define _ARM_IF_LSHIFTI(_Cond, DRR,SRR,Imm,Label)  _ARM_LSHIFTI_UCF(DRR,SRR,Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)
# define _ARM_IF_LSHIFT(_Cond, DRR,SRR1,SRR2,Label) _ARM_LSHIFT_UCF(DRR,SRR1,SRR2) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)

/* ------------------------------------------------------------------------- */

    /*
     * Compare
     */

# define _ARM_IF_CMPI(_Op, LHSRR, Imm, Label)       _ARM_CMPI(LHSRR,Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Op),Label)
# define _ARM_IF_CMP(_Op, LHSRR, RHSRR, Label)      _ARM_CMP(LHSRR,RHSRR) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Op),Label)
# define _ARM_IF_CMPNEGI(_Op, LHSRR, Imm, Label)    _ARM_CMPNEGI(LHSRR,Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Op),Label)
# define _ARM_IF_CMPNEG(_Op, LHSRR, RHSRR, Label)   _ARM_CMPNEG(LHSRR,RHSRR) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Op),Label)

/* Usage sample:

ARM:
	_ARM_IF_EQ0([tmp], L1)
		stmt1
		_ARM_GOTO_IMM(L2)
	_ARM_ENDIF(L1)
		stmt2
	L2:
	stmt3

C:
	if (0 == tmp)
		stmt1;
	else
		stmt2;
	stmt3;
 */
# define _ARM_IF_EQ0(LHSRR, Label)          _ARM_IF_CMPI(EQ, LHSRR, 0, Label)
# define _ARM_IF_NE0(LHSRR, Label)          _ARM_IF_CMPI(NE, LHSRR, 0, Label)

# define _ARM_IF_EQI(LHSRR, Imm, Label)     _ARM_IF_CMPI(EQ, LHSRR, Imm, Label)
# define _ARM_IF_NEI(LHSRR, Imm, Label)     _ARM_IF_CMPI(NE, LHSRR, Imm, Label)
# define _ARM_IF_LTI(LHSRR, Imm, Label)     _ARM_IF_CMPI(LT, LHSRR, Imm, Label)
# define _ARM_IF_LEI(LHSRR, Imm, Label)     _ARM_IF_CMPI(LE, LHSRR, Imm, Label)
# define _ARM_IF_GEI(LHSRR, Imm, Label)     _ARM_IF_CMPI(GE, LHSRR, Imm, Label)
# define _ARM_IF_GTI(LHSRR, Imm, Label)     _ARM_IF_CMPI(GT, LHSRR, Imm, Label)

# define _ARM_IF_EQ(LHSRR, RHSRR, Label)    _ARM_IF_CMP(EQ, LHSRR, RHSRR, Label)
# define _ARM_IF_NE(LHSRR, RHSRR, Label)    _ARM_IF_CMP(NE, LHSRR, RHSRR, Label)
# define _ARM_IF_LT(LHSRR, RHSRR, Label)    _ARM_IF_CMP(LT, LHSRR, RHSRR, Label)
# define _ARM_IF_LE(LHSRR, RHSRR, Label)    _ARM_IF_CMP(LE, LHSRR, RHSRR, Label)
# define _ARM_IF_GE(LHSRR, RHSRR, Label)    _ARM_IF_CMP(GE, LHSRR, RHSRR, Label)
# define _ARM_IF_GT(LHSRR, RHSRR, Label)    _ARM_IF_CMP(GT, LHSRR, RHSRR, Label)


# define _ARM_IF_EQNEG0(LHSRR, Label)       _ARM_IF_CMPNEGI(EQ, LHSRR, 0, Label)
# define _ARM_IF_NENEG0(LHSRR, Label)       _ARM_IF_CMPNEGI(NE, LHSRR, 0, Label)

# define _ARM_IF_EQNEGI(LHSRR, Imm, Label)  _ARM_IF_CMPNEGI(EQ, LHSRR, Imm, Label)
# define _ARM_IF_NENEGI(LHSRR, Imm, Label)  _ARM_IF_CMPNEGI(NE, LHSRR, Imm, Label)
# define _ARM_IF_LTNEGI(LHSRR, Imm, Label)  _ARM_IF_CMPNEGI(LT, LHSRR, Imm, Label)
# define _ARM_IF_LENEGI(LHSRR, Imm, Label)  _ARM_IF_CMPNEGI(LE, LHSRR, Imm, Label)
# define _ARM_IF_GENEGI(LHSRR, Imm, Label)  _ARM_IF_CMPNEGI(GE, LHSRR, Imm, Label)
# define _ARM_IF_GTNEGI(LHSRR, Imm, Label)  _ARM_IF_CMPNEGI(GT, LHSRR, Imm, Label)

# define _ARM_IF_EQNEG(LHSRR, RHSRR, Label) _ARM_IF_CMPNEG(EQ, LHSRR, RHSRR, Label)
# define _ARM_IF_NENEG(LHSRR, RHSRR, Label) _ARM_IF_CMPNEG(NE, LHSRR, RHSRR, Label)
# define _ARM_IF_LTNEG(LHSRR, RHSRR, Label) _ARM_IF_CMPNEG(LT, LHSRR, RHSRR, Label)
# define _ARM_IF_LENEG(LHSRR, RHSRR, Label) _ARM_IF_CMPNEG(LE, LHSRR, RHSRR, Label)
# define _ARM_IF_GENEG(LHSRR, RHSRR, Label) _ARM_IF_CMPNEG(GE, LHSRR, RHSRR, Label)
# define _ARM_IF_GTNEG(LHSRR, RHSRR, Label) _ARM_IF_CMPNEG(GT, LHSRR, RHSRR, Label)

/* ------------------------------------------------------------------------- */

    /*
     * Move
     */

# define _ARM_IF_MOVI(_Cond, DRR, Imm, Label)   _ARM_MOVI_UCF(DRR, Imm) _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)
# define _ARM_IF_MOV(_Cond, DRR, SRR, Label)    _ARM_MOV_UCF(DRR, SRR)  _ARM_GOTO_IF(_ARM_CCOND_MAP(_Cond),Label)

# define _ARM_IF_MOVI_NE0(DRR, Imm, Label)      _ARM_IF_MOVI(NE, DRR, Imm, Label)
# define _ARM_IF_MOVI_GE0(DRR, Imm, Label)      _ARM_IF_MOVI(GE, DRR, Imm, Label)
# define _ARM_IF_MOVI_GT0(DRR, Imm, Label)      _ARM_IF_MOVI(GT, DRR, Imm, Label)

# define _ARM_IF_MOV_NE0(DRR, SRR, Label)       _ARM_IF_MOV(NE, DRR, SRR, Label)
# define _ARM_IF_MOV_GE0(DRR, SRR, Label)       _ARM_IF_MOV(GE, DRR, SRR, Label)
# define _ARM_IF_MOV_GT0(DRR, SRR, Label)       _ARM_IF_MOV(GT, DRR, SRR, Label)

/* @} */


/* ========================================================================= */

/*
 * Print
 * @{
 */

# define _ARM_PRINTRR1(_Fmt,RR) \
    _ARM_PUSH_ALLRR()           \
	__ARM_PRINTRR1(_Fmt,RR)     \
	_ARM_POP_ALLRR()

# define _ARM_PRINTRR2(_Fmt,RR1,RR2) \
    _ARM_PUSH_ALLRR()                \
	__ARM_PRINTRR2(_Fmt,RR1,RR2)     \
	_ARM_POP_ALLRR()

# define _ARM_PRINTRR3(_Fmt,RR1,RR2,RR3) \
    _ARM_PUSH_ALLRR()                    \
	__ARM_PRINTRR3(_Fmt,RR1,RR2,RR3)     \
	_ARM_POP_ALLRR()

/* @} */
/* @} */


/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

/*
 * NEON
 * @{
 */


/* ************************************************************************* */

/*
 * Move operations
 * @{
 */

    /*
     * Duplicate/Set immediate value into every element of the destination vector register
	 *
	 * (dM[0],..., dM[N]) = Imm
     */
# define _ARM_VSETIn_Q(W,VRI,Imm)       _ARM_VDUPI_Q(W,VRI,Imm)
# define _ARM_VSETIn_D(W,VRI,Imm)       _ARM_VDUPI_D(W,VRI,Imm)
# define _ARM_VSETFIn_Q(W,VRI,Imm)      _ARM_VDUPFI_Q(W,VRI,Imm)
# define _ARM_VSETFIn_D(W,VRI,Imm)      _ARM_VDUPFI_D(W,VRI,Imm)
# define _ARM_VSETFIn_S(W,VRI,Imm)      _ARM_VDUPFI_S(W,VRI,Imm)
# define _ARM_VSETFI_S(VRI,Imm)         _ARM_VSETFIn_S(32,VRI,Imm)
# ifdef _ARM_VDUPIEX_Q
#  define _ARM_VSETInEX_Q(W,VRI,Imm,Ext)_ARM_VDUPIEX_Q(W,VRI,Imm,Ext)
#  define _ARM_VSETInEX_D(W,VRI,Imm,Ext)_ARM_VDUPIEX_D(W,VRI,Imm,Ext)
# endif

/* ------------------------------------------------------------------------- */

    /*
     * Duplicate/Set [regular register] scalar into every element of the destination vector register
	 *
	 * (dM[0],..., dM[N]) = $RR
     */
# define _ARM_VSETn_Q(W,DVRI,RR)        _ARM_VDUPR_Q(W,DVRI,RR)
# define _ARM_VSETn_D(W,DVRI,RR)        _ARM_VDUPR_D(W,DVRI,RR)

/* @} */


/* ========================================================================= */

/*
 * Converting/Saturating operations
 * @{
 */

/* w - source data width
 *
 * widen:   Q    --> (Q,Q) ;; uint8 --> int16, int16 --> int32, int32 --> int64
 * widenf:  Q    --> (Q,Q) ;; float16 --> float32, float32 --> float64
 * narrow: (Q,Q) -->  Q    ;; int16 --> uint8, int32 --> int16, int64 --> int32
 * narrowf:(Q,Q) -->  Q    ;; float32 --> float16, float64 --> float32
 */


# define _ARM_VWIDEN_Q2Q1(W,DVRI0,DVRI1,SVRI)   \
	_ARM_VWIDEN_Q1D1(W,DVRI0,_ARM_QDMAP0(SVRI)) \
	_ARM_VWIDEN_Q1D1(W,DVRI1,_ARM_QDMAP1(SVRI))

# define _ARM_VWIDENF_Q2Q1(W,DVRI0,DVRI1,SVRI)   \
    _ARM_VWIDENF_Q1D1(W,DVRI0,_ARM_QDMAP0(SVRI)) \
    _ARM_VWIDENF_Q1D1(W,DVRI1,_ARM_QDMAP1(SVRI))

# define _ARM_VNARROW_Q1Q2(W,DVRI,SVRI0,SVRI1)   \
	_ARM_VNARROW_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0) \
	_ARM_VNARROW_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1)

# define _ARM_VNARROWF_Q1Q2(W,DVRI,SVRI0,SVRI1)   \
    _ARM_VNARROWF_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0) \
    _ARM_VNARROWF_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1)

# ifdef _ARM_VUZP_Q
/* narrow: inplace variant; w - source data width */
#  define _ARM_VNARROW_D2(W,SDVRI0,SDVRI1)  _ARM_VUZP_D(_ARM_WDIV2(W),SDVRI0,SDVRI1)
#  define _ARM_VNARROW_Q1(W,SDVRI)          _ARM_VUZP_D(_ARM_WDIV2(W),_ARM_QDMAP0(SDVRI),_ARM_QDMAP1(SDVRI))
#  define _ARM_VNARROW_Q2(W,SDVRI0,SDVRI1)  _ARM_VUZP_Q(_ARM_WDIV2(W),SDVRI0,SDVRI1)
# endif
# ifdef _ARM_VUZP1_Q
/* narrow: inplace variant; w - source data width */
#  define _ARM_VNARROW_D2(W,SDVRI0,SDVRI1)  _ARM_VUZP1_D(_ARM_WDIV2(W),SDVRI0,SDVRI0,SDVRI1)
#  define _ARM_VNARROW_Q1(W,SDVRI)          _ARM_VUZP1_Q(_ARM_WDIV2(W),SDVRI,SDVRI,SDVRI)
#  define _ARM_VNARROW_Q2(W,SDVRI0,SDVRI1)  _ARM_VUZP1_Q(_ARM_WDIV2(W),SDVRI0,SDVRI0,SDVRI1)
# endif


/* narrow with saturate: (Q,Q) --> Q */
# define _ARM_VSATURATE_UU_Q1Q2(W,DVRI,SVRI0,SVRI1)   \
    _ARM_VSATURATE_UU_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0) \
    _ARM_VSATURATE_UU_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1)

# define _ARM_VSATURATE_SS_Q1Q2(W,DVRI,SVRI0,SVRI1)   \
    _ARM_VSATURATE_SS_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0) \
    _ARM_VSATURATE_SS_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1)

# define _ARM_VSATURATE_SU_Q1Q2(W,DVRI,SVRI0,SVRI1)   \
    _ARM_VSATURATE_SU_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0) \
    _ARM_VSATURATE_SU_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1)

/* ------------------------------------------------------------------------- */

    /*
     * vector convert signed/unsigned integer to float
     */
# define _ARM_VSITOF_Q(DVRI,SVRI)           _ARM_VSITOFEX_Q(32,DVRI,SVRI)
# define _ARM_VSITOF_D(DVRI,SVRI)           _ARM_VSITOFEX_D(32,DVRI,SVRI)
# define _ARM_VSITOF_S(DVRI,SVRI)           _ARM_VSITOFEX_S(32,DVRI,SVRI)
# define _ARM_VUITOF_Q(DVRI,SVRI)           _ARM_VUITOFEX_Q(32,DVRI,SVRI)
# define _ARM_VUITOF_D(DVRI,SVRI)           _ARM_VUITOFEX_D(32,DVRI,SVRI)
# define _ARM_VUITOF_S(DVRI,SVRI)           _ARM_VUITOFEX_S(32,DVRI,SVRI)

/* ------------------------------------------------------------------------- */

    /*
     * vector convert truncate float to signed/unsigned integer
     */
# define _ARM_VFTOSI_S(DVRI,SVRI)           _ARM_RND_VFTOSIEX_S(32,DVRI,SVRI,Z)
# define _ARM_VFTOSI_D(DVRI,SVRI)           _ARM_RND_VFTOSIEX_D(32,DVRI,SVRI,Z)
# define _ARM_VFTOSI_Q(DVRI,SVRI)           _ARM_RND_VFTOSIEX_Q(32,DVRI,SVRI,Z)
# define _ARM_VFTOUI_S(DVRI,SVRI)           _ARM_RND_VFTOUIEX_S(32,DVRI,SVRI,Z)
# define _ARM_VFTOUI_D(DVRI,SVRI)           _ARM_RND_VFTOUIEX_D(32,DVRI,SVRI,Z)
# define _ARM_VFTOUI_Q(DVRI,SVRI)           _ARM_RND_VFTOUIEX_Q(32,DVRI,SVRI,Z)

    /*
     * vector convert truncate double to signed/unsigned integer
     */
/* D --> S */
# ifdef _ARM_RND_VF64TOSIEX_S1D1
#  define _ARM_VF64TOSI_S1D1(DVRI,SVRI)         _ARM_RND_VF64TOSIEX_S1D1(DVRI,SVRI,Z)
# endif
/* Q --> D */
# define _ARM_VF64TOSI_D1Q1(DVRI,SVRI)          _ARM_RND_VF64TOSIEX_D1Q1(DVRI,SVRI,Z)
/* (Q,Q) --> Q */
# define _ARM_VF64TOSI_Q1Q2(DVRI,SVRI0,SVRI1)   _ARM_RND_VF64TOSIEX_Q1Q2(DVRI,SVRI0,SVRI1,Z)
/* D --> S */
# ifdef _ARM_RND_VF64TOUIEX_S1D1
#  define _ARM_VF64TOUI_S1D1(DVRI,SVRI)         _ARM_RND_VF64TOUIEX_S1D1(DVRI,SVRI,Z)
# endif
/* Q --> D */
# define _ARM_VF64TOUI_D1Q1(DVRI,SVRI)          _ARM_RND_VF64TOUIEX_D1Q1(DVRI,SVRI,Z)
/* (Q,Q) --> Q */
# define _ARM_VF64TOUI_Q1Q2(DVRI,SVRI0,SVRI1)   _ARM_RND_VF64TOUIEX_Q1Q2(DVRI,SVRI0,SVRI1,Z)

    /*
     * vector convert truncate float to signed/unsigned [long] integer
     */
# define _ARM_VFTOSI32(D,S)                 _ARM_RND_VFTOSIEX(32,32,D,S,Z)
# define _ARM_VFTOUI32(D,S)                 _ARM_RND_VFTOUIEX(32,32,D,S,Z)
# define _ARM_VFTOSI64(D,S)                 _ARM_RND_VFTOSIEX(64,32,D,S,Z)
# define _ARM_VFTOUI64(D,S)                 _ARM_RND_VFTOUIEX(64,32,D,S,Z)
# define _ARM_VFTOSI(D,S)                   _ARM_VFTOSI32(D,S)
# define _ARM_VFTOUI(D,S)                   _ARM_VFTOUI32(D,S)
# define _ARM_VFTOLI(D,S)                   _ARM_VFTOSI64(D,S)
# define _ARM_VFTOLU(D,S)                   _ARM_VFTOUI64(D,S)

    /*
     * vector convert truncate double to signed/unsigned [long] integer
     */
# define _ARM_VF64TOSI32(D,S)               _ARM_RND_VFTOSIEX(32,64,D,S,Z)
# define _ARM_VF64TOUI32(D,S)               _ARM_RND_VFTOUIEX(32,64,D,S,Z)
# define _ARM_VF64TOSI64(D,S)               _ARM_RND_VFTOSIEX(64,64,D,S,Z)
# define _ARM_VF64TOUI64(D,S)               _ARM_RND_VFTOUIEX(64,64,D,S,Z)
# define _ARM_VF64TOSI(D,S)                 _ARM_VF64TOSI32(D,S)
# define _ARM_VF64TOUI(D,S)                 _ARM_VF64TOUI32(D,S)
# define _ARM_VF64TOLI(D,S)                 _ARM_VF64TOSI64(D,S)
# define _ARM_VF64TOLU(D,S)                 _ARM_VF64TOUI64(D,S)

/* ------------------------------------------------------------------------- */

    /*
     * vector rounding convert float to signed/unsigned integer
     */
# define _ARM_RND_VFTOSI_S(DVRI,SVRI)       _ARM_RND_VFTOSIEX_S(32,DVRI,SVRI,A)
# define _ARM_RND_VFTOUI_S(DVRI,SVRI)       _ARM_RND_VFTOUIEX_S(32,DVRI,SVRI,A)

    /*
     * vector rounding convert double to signed/unsigned integer
     */
/* D --> S */
# ifdef _ARM_RND_VF64TOSIEX_S1D1
#  define _ARM_RND_VF64TOSI_S1D1(DVRI,SVRI)         _ARM_RND_VF64TOSIEX_S1D1(DVRI,SVRI,A)
# endif
/* Q --> D */
# define _ARM_RND_VF64TOSI_D1Q1(DVRI,SVRI)          _ARM_RND_VF64TOSIEX_D1Q1(DVRI,SVRI,A)
/* (Q,Q) --> Q */
# define _ARM_RND_VF64TOSI_Q1Q2(DVRI,SVRI0,SVRI1)   _ARM_RND_VF64TOSIEX_Q1Q2(DVRI,SVRI0,SVRI1,A)
/* D --> S */
# ifdef _ARM_RND_VF64TOUIEX_S1D1
#  define _ARM_RND_VF64TOUI_S1D1(DVRI,SVRI)         _ARM_RND_VF64TOUIEX_S1D1(DVRI,SVRI,A)
# endif
/* Q --> D */
# define _ARM_RND_VF64TOUI_D1Q1(DVRI,SVRI)          _ARM_RND_VF64TOUIEX_D1Q1(DVRI,SVRI,A)
/* (Q,Q) --> Q */
# define _ARM_RND_VF64TOUI_Q1Q2(DVRI,SVRI0,SVRI1)   _ARM_RND_VF64TOUIEX_Q1Q2(DVRI,SVRI0,SVRI1,A)

    /*
     * vector rounding convert float to signed/unsigned [long] integer
     */
# define _ARM_RND_VFTOSI32(D,S)             _ARM_RND_VFTOSIEX(32,32,D,S,N)
# define _ARM_RND_VFTOUI32(D,S)             _ARM_RND_VFTOUIEX(32,32,D,S,N)
# define _ARM_RND_VFTOSI64(D,S)             _ARM_RND_VFTOSIEX(64,32,D,S,N)
# define _ARM_RND_VFTOUI64(D,S)             _ARM_RND_VFTOUIEX(64,32,D,S,N)
# define _ARM_RND_VFTOSI(D,S)               _ARM_RND_VFTOSI32(D,S)
# define _ARM_RND_VFTOUI(D,S)               _ARM_RND_VFTOUI32(D,S)
# define _ARM_RND_VFTOLI(D,S)               _ARM_RND_VFTOSI64(D,S)
# define _ARM_RND_VFTOLU(D,S)               _ARM_RND_VFTOUI64(D,S)

    /*
     * vector rounding convert double to signed/unsigned [long] integer
     */
# define _ARM_RND_VF64TOSI32(D,S)           _ARM_RND_VFTOSIEX(32,64,D,S,N)
# define _ARM_RND_VF64TOUI32(D,S)           _ARM_RND_VFTOUIEX(32,64,D,S,N)
# define _ARM_RND_VF64TOSI64(D,S)           _ARM_RND_VFTOSIEX(64,64,D,S,N)
# define _ARM_RND_VF64TOUI64(D,S)           _ARM_RND_VFTOUIEX(64,64,D,S,N)
# define _ARM_RND_VF64TOSI(D,S)             _ARM_RND_VF64TOSI32(D,S)
# define _ARM_RND_VF64TOUI(D,S)             _ARM_RND_VF64TOUI32(D,S)
# define _ARM_RND_VF64TOLI(D,S)             _ARM_RND_VF64TOSI64(D,S)
# define _ARM_RND_VF64TOLU(D,S)             _ARM_RND_VF64TOUI64(D,S)

/* ------------------------------------------------------------------------- */

	/*
     * vector convert float to double
	 */
/* S --> D */
# define _ARM_VF32TOF64_D1S1(DVRI,SVRI)         _ARM_VWIDENF_D1S1(32,DVRI,SVRI)
/* D --> Q */
# define _ARM_VF32TOF64_Q1D1(DVRI,SVRI)         _ARM_VWIDENF_Q1D1(32,DVRI,SVRI)
/* Q --> (Q,Q) */
# define _ARM_VF32TOF64_Q2Q1(DVRI0,DVRI1,SVRI)  _ARM_VWIDENF_Q2Q1(32,DVRI0,DVRI1,SVRI)


	/*
     * vector convert double to float
	 */
/* D --> S */
# define _ARM_VF64TOF32_S1D1(DVRI,SVRI)         _ARM_VNARROWF_S1D1(64,DVRI,SVRI)
/* Q --> D */
# define _ARM_VF64TOF32_D1Q1(DVRI,SVRI)         _ARM_VNARROWF_D1Q1(64,DVRI,SVRI)
/* (Q,Q) --> Q */
# define _ARM_VF64TOF32_Q1Q2(DVRI,SVRI0,SVRI1)  _ARM_VNARROWF_Q1Q2(64,DVRI,SVRI0,SVRI1)

/* @} */


/* ========================================================================= */

/*
 * Generic arithmetic operations
 * @{
 *
 * :w  - data width
 * :ww - double data width
 * for instance: w = 8, ww = 16
 *               long operation: ww = w + w
 *               wide operation: ww = ww + w
 *
 * vector negate               : r[i]:w   = -a[i]:w
 *
 * vector add                           : r[i]:w   = a[i]:w + b[i]:w
 * vector add saturated                 : r[i]:w   = saturate(a[i]:w + b[i]:w)
 * vector add long                      : r[i]:ww  = a[i]:w + b[i]:w
 * vector add wide                      : r[i]:ww  = a[i]:ww + b[i]:w
 * vector halving add                   : r[i]:w   = (a[i]:w + b[i]:w) >> 1
 * vector rounding halving add          : r[i]:w   = (a[i]:w + b[i]:w + 1) >> 1
 *
 * vector sub                           : r[i]:w   = a[i]:w - b[i]:w
 * vector sub saturated                 : r[i]:w   = saturate(a[i]:w - b[i]:w)
 * vector sub long                      : r[i]:ww  = a[i]:w - b[i]:w
 * vector sub wide                      : r[i]:ww  = a[i]:ww - b[i]:w
 * vector halving sub                   : r[i]:w   = (a[i]:w - b[i]:w) >> 1
 *
 * vector mul                           : r[i]:w   = a[i]:w * b[i]:w
 * vector mul long                      : r[i]:ww  = a[i]:w * b[i]:w
 * vector mul long saturated            : r[i]:ww  = saturate(a[i]:w * b[i]:w)
 * vector mul accumulate                : r[i]:w  += a[i]:w * b[i]:w
 * vector mul accumulate long           : r[i]:ww += a[i]:w * b[i]:w
 * vector mul accumulate saturated long : r[i]:ww = saturate(r[i]:ww + (a[i]:w * b[i]:w))
 * vector mul subtract                  : r[i]:w  -= a[i]:w * b[i]:w
 * vector mul subtract long             : r[i]:ww -= a[i]:w * b[i]:w
 * vector mul subtract saturated long   : r[i]:ww = saturate(r[i]:ww - (a[i]:w * b[i]:w))
 *
 * vector doubling mul saturated with right shift 16/32          : r = saturate((a * b * 2) >>{U} 16{32})
 * vector rounding doubling mul saturated with right shift 16/32 : r = saturate(round((a * b * 2) >>{U} 16{32}))
 */

/* ------------------------------------------------------------------------- */

    /*
     * VADD
     */

/* add long: (Q,Q) --> (Q,Q) */
# define _ARM_VADD_WIDEN_Q2Q1Q1(W, DVRI0,DVRI1,SVRI1,SVRI2)                \
    _ARM_VADD_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),_ARM_QDMAP0(SVRI2)) \
    _ARM_VADD_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),_ARM_QDMAP1(SVRI2))

/* add wide: (Q,Q,Q) --> (Q,Q) */
# define _ARM_VADD_WIDEN_Q2Q2Q1(W, DVRI0,DVRI1,SVRI0,SVRI1,SVRI2) \
    _ARM_VADD_WIDEN_Q1Q1D1(W, DVRI0,SVRI0,_ARM_QDMAP0(SVRI2))     \
    _ARM_VADD_WIDEN_Q1Q1D1(W, DVRI1,SVRI1,_ARM_QDMAP1(SVRI2))


/* add narrow, selecting high half */
/* narrow: ((Q,Q),(Q,Q)) --> Q */
# define _ARM_VADDH_NARROW_Q1Q2Q2(W, DVRI,SVRI10,SVRI11,SVRI20,SVRI21) \
    _ARM_VADDH_NARROW_D1Q1Q1(W,_ARM_QDMAP0(DVRI),SVRI10,SVRI20) \
    _ARM_VADDH_NARROW_D1Q1Q1(W,_ARM_QDMAP1(DVRI),SVRI11,SVRI21)

/* rounding add narrow, selecting high half */
/* narrow: ((Q,Q),(Q,Q)) --> Q */
# define _ARM_RND_VADDH_NARROW_Q1Q2Q2(W, DVRI,SVRI10,SVRI11,SVRI20,SVRI21) \
    _ARM_RND_VADDH_NARROW_D1Q1Q1(W,_ARM_QDMAP0(DVRI),SVRI10,SVRI20) \
    _ARM_RND_VADDH_NARROW_D1Q1Q1(W,_ARM_QDMAP1(DVRI),SVRI11,SVRI21)

/* ------------------------------------------------------------------------- */

    /*
     * VSUB
     */

/* sub long: (Q,Q) --> (Q,Q) */
# define _ARM_VSUB_WIDEN_Q2Q1Q1(W, DVRI0,DVRI1,SVRI1,SVRI2)                \
    _ARM_VSUB_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),_ARM_QDMAP0(SVRI2)) \
    _ARM_VSUB_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),_ARM_QDMAP1(SVRI2))

/* sub wide: (Q,Q,Q) --> (Q,Q) */
# define _ARM_VSUB_WIDEN_Q2Q2Q1(W, DVRI0,DVRI1,SVRI0,SVRI1,SVRI2) \
    _ARM_VSUB_WIDEN_Q1Q1D1(W, DVRI0,SVRI0,_ARM_QDMAP0(SVRI2))     \
    _ARM_VSUB_WIDEN_Q1Q1D1(W, DVRI1,SVRI1,_ARM_QDMAP1(SVRI2))


/* sub narrow, selecting high half */
/* narrow: ((Q,Q),(Q,Q)) --> Q */
# define _ARM_VSUBH_NARROW_Q1Q2Q2(W, DVRI,SVRI10,SVRI11,SVRI20,SVRI21) \
    _ARM_VSUBH_NARROW_D1Q1Q1(W,_ARM_QDMAP0(DVRI),SVRI10,SVRI20) \
    _ARM_VSUBH_NARROW_D1Q1Q1(W,_ARM_QDMAP1(DVRI),SVRI11,SVRI21)

/* rounding sub narrow, selecting high half */
/* narrow: ((Q,Q),(Q,Q)) --> Q */
# define _ARM_RND_VSUBH_NARROW_Q1Q2Q2(W, DVRI,SVRI10,SVRI11,SVRI20,SVRI21) \
    _ARM_RND_VSUBH_NARROW_D1Q1Q1(W,_ARM_QDMAP0(DVRI),SVRI10,SVRI20) \
    _ARM_RND_VSUBH_NARROW_D1Q1Q1(W,_ARM_QDMAP1(DVRI),SVRI11,SVRI21)

/* ------------------------------------------------------------------------- */

    /*
     * VMUL
     */

/* mul long: (Q,Q) --> (Q,Q)
   mulS long: (Q,D[L]) --> (Q,Q) */
# define _ARM_VMUL_WIDEN_Q2Q1Q1(W, DVRI0,DVRI1,SVRI1,SVRI2)                \
    _ARM_VMUL_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),_ARM_QDMAP0(SVRI2)) \
    _ARM_VMUL_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),_ARM_QDMAP1(SVRI2))
# define _ARM_VMULS_WIDEN_Q2Q1D1(W, DVRI0,DVRI1,SVRI1,L,SVRI2)   \
    _ARM_VMULS_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),L,SVRI2) \
    _ARM_VMULS_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),L,SVRI2)

/* mul wide: (Q,Q,Q) --> (Q,Q)
   mulS wide: (Q,Q,D[L]) --> (Q,Q) */
# define _ARM_VMUL_WIDEN_Q2Q2Q1(W, DVRI0,DVRI1,SVRI0,SVRI1,SVRI2) \
    _ARM_VMUL_WIDEN_Q1Q1D1(W, DVRI0,SVRI0,_ARM_QDMAP0(SVRI2))     \
    _ARM_VMUL_WIDEN_Q1Q1D1(W, DVRI1,SVRI1,_ARM_QDMAP1(SVRI2))
# define _ARM_VMULS_WIDEN_Q2Q2D1(W, DVRI0,DVRI1,SVRI0,SVRI1,L,SVRI2) \
    _ARM_VMULS_WIDEN_Q1Q1D1(W, DVRI0,SVRI0,L,SVRI2)                  \
    _ARM_VMULS_WIDEN_Q1Q1D1(W, DVRI1,SVRI1,L,SVRI2)


/* doubling mul saturated long: (Q,Q) --> (Q,Q); domain: {int16, int32} */
# define _ARM_SAT_VMUL2_WIDEN_Q2Q1Q1(W, DVRI0,DVRI1,SVRI1,SVRI2)                \
    _ARM_SAT_VMUL2_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),_ARM_QDMAP0(SVRI2)) \
    _ARM_SAT_VMUL2_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),_ARM_QDMAP1(SVRI2))


/* mul accumulate/subtract long: (Q,Q) --> (Q,Q)
   mulS accumulate/subtract long: (Q,D[L]) --> (Q,Q) */

# define _ARM_VMUL_ACCUM_WIDEN_Q2Q1Q1(W, DVRI0,DVRI1,SVRI1,SVRI2)                \
    _ARM_VMUL_ACCUM_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),_ARM_QDMAP0(SVRI2)) \
    _ARM_VMUL_ACCUM_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),_ARM_QDMAP1(SVRI2))
# define _ARM_VMULS_ACCUM_WIDEN_Q2Q1D1(W, DVRI0,DVRI1,SVRI1,L,SVRI2)   \
    _ARM_VMULS_ACCUM_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),L,SVRI2) \
    _ARM_VMULS_ACCUM_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),L,SVRI2)

# define _ARM_VMUL_SUBTR_WIDEN_Q2Q1Q1(W, DVRI0,DVRI1,SVRI1,SVRI2)                \
    _ARM_VMUL_SUBTR_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),_ARM_QDMAP0(SVRI2)) \
    _ARM_VMUL_SUBTR_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),_ARM_QDMAP1(SVRI2))
# define _ARM_VMULS_SUBTR_WIDEN_Q2Q1D1(W, DVRI0,DVRI1,SVRI1,L,SVRI2)   \
    _ARM_VMULS_SUBTR_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),L,SVRI2) \
    _ARM_VMULS_SUBTR_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),L,SVRI2)


/* doubling mul accumulate/subtract saturated long: (Q,Q) --> (Q,Q); domain: {int16, int32} */

# define _ARM_SAT_VMUL2_ACCUM_WIDEN_Q2Q1Q1(W, DVRI0,DVRI1,SVRI1,SVRI2)                \
    _ARM_SAT_VMUL2_ACCUM_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),_ARM_QDMAP0(SVRI2)) \
    _ARM_SAT_VMUL2_ACCUM_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),_ARM_QDMAP1(SVRI2))

# define _ARM_SAT_VMUL2_SUBTR_WIDEN_Q2Q1Q1(W, DVRI0,DVRI1,SVRI1,SVRI2)                \
    _ARM_SAT_VMUL2_SUBTR_WIDEN_Q1D1D1(W, DVRI0,_ARM_QDMAP0(SVRI1),_ARM_QDMAP0(SVRI2)) \
    _ARM_SAT_VMUL2_SUBTR_WIDEN_Q1D1D1(W, DVRI1,_ARM_QDMAP1(SVRI1),_ARM_QDMAP1(SVRI2))


/* A,X,Y - named variables [gcc] inline assembler */
# define _ARM_VFMA(A,X,Y)                   _ARM_VFMAEX(32,A,X,Y)
# define _ARM_VFMS(A,X,Y)                   _ARM_VFMSEX(32,A,X,Y)
# define _ARM_VFMA64(A,X,Y)                 _ARM_VFMAEX(64,A,X,Y)
# define _ARM_VFMS64(A,X,Y)                 _ARM_VFMSEX(64,A,X,Y)

/* @} */


/* ========================================================================= */

/*
 * Logical bitwise operations
 * @{
 *
 * vector not
 * vector and
 * vector and complement: BIC
 * vector or
 * vector or complement: ORN
 * vector xor
 * vector set all bits
 * vector clear all bits
 * vector bit select           : bit(i, dst) = (bit(i, dst) ? bit(i, src1) : bit(i, src2))
 */

    /*
     * NOT
     */

# define _ARM_VNOT_Q(W, DVRI,SVRI)      _ARM_VMVN_Q(DVRI,SVRI)
# define _ARM_VNOT_D(W, DVRI,SVRI)      _ARM_VMVN_D(DVRI,SVRI)

/* @} */


/* ========================================================================= */

/*
 * Shift operations
 * @{
 *
 * vector shift left by immediate
 * vector shift left saturated by immediate
 * vector shift left long by immediate
 *
 * (negative shift-value shift right)
 * vector shift left by signed var
 * vector rounding shift left by signed
 * vector shift left saturated by signed
 * vector rounding shift left saturated by signed
 *
 * vector [rounding] shift right by immediate
 * vector [rounding] shift right narrow by immediate
 * vector saturating [rounding] shift right narrow by immediate
 */

    /*
     * VSHL by immediate
     */

/*
 * shift long
 */

# define _ARM_VSHLI_WIDEN_Q2Q1(W, DVRI0,DVRI1,SVRI,Imm)   \
	_ARM_VSHLI_WIDEN_Q1D1(W, DVRI0,_ARM_QDMAP0(SVRI),Imm) \
	_ARM_VSHLI_WIDEN_Q1D1(W, DVRI1,_ARM_QDMAP1(SVRI),Imm)

/* signed shift long */
# define _ARM_VSHLSI_WIDEN_Q2Q1(W, DVRI0,DVRI1,SVRI,Imm)   \
	_ARM_VSHLSI_WIDEN_Q1D1(W, DVRI0,_ARM_QDMAP0(SVRI),Imm) \
	_ARM_VSHLSI_WIDEN_Q1D1(W, DVRI1,_ARM_QDMAP1(SVRI),Imm)

/* unsigned shift long */
# define _ARM_VSHLUI_WIDEN_Q2Q1(W, DVRI0,DVRI1,SVRI,Imm)   \
	_ARM_VSHLUI_WIDEN_Q1D1(W, DVRI0,_ARM_QDMAP0(SVRI),Imm) \
	_ARM_VSHLUI_WIDEN_Q1D1(W, DVRI1,_ARM_QDMAP1(SVRI),Imm)

/* ------------------------------------------------------------------------- */

    /*
     * VSHR by immediate
     */

/* [signed/unsigned] shift right narrow */
/* narrow: (Q,Q) --> Q */
# define _ARM_VSHRI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)  \
    _ARM_VSHRI_NARROW_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0,Imm) \
    _ARM_VSHRI_NARROW_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1,Imm)
//# define _ARM_VSHRSI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm) _ARM_VSHRI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)
//# define _ARM_VSHRUI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm) _ARM_VSHRI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)


/* [signed/unsigned] rounding shift right narrow */
/* narrow: (Q,Q) --> Q */
# define _ARM_RND_VSHRI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)  \
    _ARM_RND_VSHRI_NARROW_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0,Imm) \
    _ARM_RND_VSHRI_NARROW_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1,Imm)
//# define _ARM_RND_VSHRSI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm) _ARM_RND_VSHRI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)
//# define _ARM_RND_VSHRUI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm) _ARM_RND_VSHRI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)


/* [signed/unsigned] shift right narrow saturated */
/* narrow: (Q,Q) --> Q */
# define _ARM_SAT_VSHRI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)  \
    _ARM_SAT_VSHRI_NARROW_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0,Imm) \
    _ARM_SAT_VSHRI_NARROW_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1,Imm)
# define _ARM_SAT_VSHRSI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)  \
    _ARM_SAT_VSHRSI_NARROW_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0,Imm) \
    _ARM_SAT_VSHRSI_NARROW_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1,Imm)
# define _ARM_SAT_VSHRUI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)  \
    _ARM_SAT_VSHRUI_NARROW_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0,Imm) \
    _ARM_SAT_VSHRUI_NARROW_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1,Imm)


/* [signed/unsigned] rounding shift right narrow saturated */
/* narrow: (Q,Q) --> Q */
# define _ARM_SAT_RND_VSHRI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)  \
    _ARM_SAT_RND_VSHRI_NARROW_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0,Imm) \
    _ARM_SAT_RND_VSHRI_NARROW_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1,Imm)
# define _ARM_SAT_RND_VSHRSI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)  \
    _ARM_SAT_RND_VSHRSI_NARROW_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0,Imm) \
    _ARM_SAT_RND_VSHRSI_NARROW_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1,Imm)
# define _ARM_SAT_RND_VSHRUI_NARROW_Q1Q2(W, DVRI,SVRI0,SVRI1,Imm)  \
    _ARM_SAT_RND_VSHRUI_NARROW_D1Q1(W,_ARM_QDMAP0(DVRI),SVRI0,Imm) \
    _ARM_SAT_RND_VSHRUI_NARROW_D1Q1(W,_ARM_QDMAP1(DVRI),SVRI1,Imm)

/* @} */


/* ========================================================================= */

/*
 * Generic data processing operations
 * @{
 *
 * vector element shift left by immediate
 * vector element shift right by immediate
 * vector element rotate left by immediate
 * vector element rotate right by immediate
 *
 * interleaves the elements of two vector (zip)
 * de-interleaves the elements of two vector (unzip)
 */

/*
 * vector elements: {0,1,2,3,4,5,6,7}
 *
 * shift-left(r,s1,s2,3):
 *    s1: {0,1,2,3,4,5,6,7};
 *    s2: {8,9,10,11,12,13,14,15}
 *    r : {3,4,5,6,7,8,9,10}
 *
 * shift-right(r,s1,s2,3):
 *    s1: {0,1,2,3,4,5,6,7};
 *    s2: {8,9,10,11,12,13,14,15}
 *    r : {5,6,7,8,9,10,11,12}
 */
# define _ARM_VESHL_Q(W,DVRI,SVRI,SVRI2,Imm)  _ARM_VEXTRACT_Q(W,DVRI,SVRI,SVRI2,Imm)
# define _ARM_VESHL_D(W,DVRI,SVRI,SVRI2,Imm)  _ARM_VEXTRACT_D(W,DVRI,SVRI,SVRI2,Imm)
# define _ARM_VESHR_Q(W,DVRI,SVRI2,SVRI,Imm)  _ARM_VEXTRACT_Q(W,DVRI,SVRI2,SVRI,_ARM_VESHRNUM_Q_MAP(W,Imm))
# define _ARM_VESHR_D(W,DVRI,SVRI2,SVRI,Imm)  _ARM_VEXTRACT_D(W,DVRI,SVRI2,SVRI,_ARM_VESHRNUM_D_MAP(W,Imm))

# define _ARM_VEROTL_D(W,DVRI,SVRI,Imm)       _ARM_VEXTRACT_D(W,DVRI,SVRI,SVRI,Imm)
# define _ARM_VEROTL_Q(W,DVRI,SVRI,Imm)       _ARM_VEXTRACT_Q(W,DVRI,SVRI,SVRI,Imm)
# define _ARM_VEROTR_Q(W,DVRI,SVRI,Imm)       _ARM_VEXTRACT_Q(W,DVRI,SVRI,SVRI,_ARM_VESHRNUM_Q_MAP(W,Imm))
# define _ARM_VEROTR_D(W,DVRI,SVRI,Imm)       _ARM_VEXTRACT_D(W,DVRI,SVRI,SVRI,_ARM_VESHRNUM_D_MAP(W,Imm))

/* @} */


/* ========================================================================= */

/*
 * Memory access
 * @{
 *
 * vector load multiple N-element structures into one or more registers (with de-interleaving when N>1)
 * vector store multiple N-element structures from one or more registers (with interleaving when N>1)
 *
 * vector load single N-element structure into one or more registers (with de-interleaving when N>1)
 * (Multiple copies of one n-element structure are loaded into one ore more registers).
 */


/*
 * Load/Store multiple N-element structures
 */

	/*
	 * With auto-increment
	 * Load 2-element structures from memory (PRR: base address) into vector registers.
	 * Store 2-element structures to memory (PRR: base address) from vector registers.
	 *
	 * PRR register updated to (PRR + number of bytes transferred) after all the loads/stores
	 * have taken place.
	 *
	 * (VRI0, ...): (Dn,Dn+1) or (Dn,Dn+2) or (Dn,Dn+1,Dn+2,Dn+3)
	 * (VRI0, ...): (Qn,Qn+1) or (Qn,Qn+1,Qn+2,Qn+3)
	 */

# define _ARM_VLD2_Q4_AIA(W, VRI0,VRI1,VRI2,VRI3,PRR) \
	_ARM_VLD2_Q2_AIA(W, VRI0,VRI1,PRR) \
	_ARM_VLD2_Q2_AIA(W, VRI2,VRI3,PRR)

# define _ARM_VST2_Q4_AIA(W, VRI0,VRI1,VRI2,VRI3,PRR) \
	_ARM_VST2_Q2_AIA(W, VRI0,VRI1,PRR) \
	_ARM_VST2_Q2_AIA(W, VRI2,VRI3,PRR)

/* ========================================================================= */

/*
 * Load (with duplicate) single N-element structure to all lanes (into all elements of vector registers)
 */

	/*
	 * With auto-increment
	 * Load 1-element structure from memory (PRR: base address) into vector registers.
	 *
	 * PRR register updated to (PRR + number of bytes transferred) after all the loads/stores
	 * have taken place.
	 *
	 * (VRI0, ...): (Dn,Dn+1,Dn+2) or (Dn,Dn+1,Dn+2,Dn+3)
	 * (VRI0, ...): (Qn,Qn+1) or (Qn,Qn+1,Qn+2) or (Qn,Qn+1,Qn+2,Qn+3)
	 */

# define _ARM_VLD1_DUP_D3_AIA(W, VRI0,VRI1,VRI2,PRR) \
	_ARM_VLD1_DUP_D2(W, VRI0,VRI1,PRR)               \
	_ARM_VLD1_DUP_D1_AIA(W, VRI2,PRR)

# define _ARM_VLD1_DUP_D4_AIA(W, VRI0,VRI1,VRI2,VRI3,PRR) \
	_ARM_VLD1_DUP_D2(W, VRI0,VRI1,PRR)                    \
	_ARM_VLD1_DUP_D2_AIA(W, VRI2,VRI3,PRR)


# define _ARM_VLD1_DUP_Q2_AIA(W, VRI0,VRI1,PRR) \
	_ARM_VLD1_DUP_Q1(W, VRI0,PRR)               \
	_ARM_VLD1_DUP_Q1_AIA(W, VRI1,PRR)

# define _ARM_VLD1_DUP_Q3_AIA(W, VRI0,VRI1,VRI2,PRR) \
	_ARM_VLD1_DUP_Q1(W, VRI0,PRR)                    \
	_ARM_VLD1_DUP_Q1(W, VRI1,PRR)                    \
	_ARM_VLD1_DUP_Q1_AIA(W, VRI2,PRR)

# define _ARM_VLD1_DUP_Q4_AIA(W, VRI0,VRI1,VRI2,VRI3,PRR) \
	_ARM_VLD1_DUP_Q1(W, VRI0,PRR)                         \
	_ARM_VLD1_DUP_Q1(W, VRI1,PRR)                         \
	_ARM_VLD1_DUP_Q2_AIA(W, VRI2,VRI3,PRR)

/* ------------------------------------------------------------------------- */

	/*
	 * With auto-increment
	 * Load 2-element structures from memory (PRR: base address) into vector registers.
	 *
	 * PRR register updated to (PRR + number of bytes transferred) after all the loads/stores
	 * have taken place.
	 *
	 * (VRI0, ...): (Dn,Dn+1) or (Dn,Dn+2) or (Dn,Dn+1,Dn+2,Dn+3)
	 * (VRI0, ...): (Qn,Qn+1) or (Qn,Qn+1,Qn+2,Qn+3)
	 */

# define _ARM_VLD2_DUP_D4_AIA(W, VRI0,VRI1,VRI2,VRI3,PRR) \
	_ARM_VLD2_DUP_D2(W, VRI0,VRI1,PRR)                    \
	_ARM_VLD2_DUP_D2_AIA(W, VRI2,VRI3,PRR)

# define _ARM_VLD2_DUP_Q4_AIA(W, VRI0,VRI1,VRI2,VRI3,PRR) \
	_ARM_VLD2_DUP_Q2(W, VRI0,VRI1,PRR)                    \
	_ARM_VLD2_DUP_Q2_AIA(W, VRI2,VRI3,PRR)

/* ========================================================================= */

/*
 * Load/Store single N-element structure to/from one lane
 */

/* @{ */
# define _ARM_VLD1_XLANE_D1(W, L,VRI,PRR)       VIPM_GLUE(_ARM_VLD1_XLANE_D1_,W)(W, L,VRI,PRR)
# define _ARM_VST1_XLANE_D1(W, L,VRI,PRR)       VIPM_GLUE(_ARM_VST1_XLANE_D1_,W)(W, L,VRI,PRR)
# define _ARM_VLD1_XLANE_D1_AIA(W, L,VRI,PRR)   VIPM_GLUE(_ARM_VLD1_XLANE_D1_AIA_,W)(W, L,VRI,PRR)
# define _ARM_VST1_XLANE_D1_AIA(W, L,VRI,PRR)   VIPM_GLUE(_ARM_VST1_XLANE_D1_AIA_,W)(W, L,VRI,PRR)
# define _ARM_VLD1_XLANE_D1_IA(W,L,VRI,PRR,ORR) VIPM_GLUE(_ARM_VLD1_XLANE_D1_IA_,W)(W, L,VRI,PRR,ORR)
# define _ARM_VST1_XLANE_D1_IA(W,L,VRI,PRR,ORR) VIPM_GLUE(_ARM_VST1_XLANE_D1_IA_,W)(W, L,VRI,PRR,ORR)


# define _ARM_VLD1_XLANE_D1_8(W, L,VRI,PRR)         _ARM_VLD1_LANE_D1(W, L,VRI,PRR)
# define _ARM_VLD1_XLANE_D1_16(W, L,VRI,PRR)        _ARM_VLD1_LANE_D1(W, L,VRI,PRR)
# define _ARM_VLD1_XLANE_D1_32(W, L,VRI,PRR)        _ARM_VLD1_LANE_D1(W, L,VRI,PRR)
# define _ARM_VLD1_XLANE_D1_64(W, L,VRI,PRR)        _ARM_VLD1_D1(W, VRI,PRR)
# define _ARM_VST1_XLANE_D1_8(W, L,VRI,PRR)         _ARM_VST1_LANE_D1(W, L,VRI,PRR)
# define _ARM_VST1_XLANE_D1_16(W, L,VRI,PRR)        _ARM_VST1_LANE_D1(W, L,VRI,PRR)
# define _ARM_VST1_XLANE_D1_32(W, L,VRI,PRR)        _ARM_VST1_LANE_D1(W, L,VRI,PRR)
# define _ARM_VST1_XLANE_D1_64(W, L,VRI,PRR)        _ARM_VST1_D1(W, VRI,PRR)

# define _ARM_VLD1_XLANE_D1_AIA_8(W, L,VRI,PRR)     _ARM_VLD1_LANE_D1_AIA(W, L,VRI,PRR)
# define _ARM_VLD1_XLANE_D1_AIA_16(W, L,VRI,PRR)    _ARM_VLD1_LANE_D1_AIA(W, L,VRI,PRR)
# define _ARM_VLD1_XLANE_D1_AIA_32(W, L,VRI,PRR)    _ARM_VLD1_LANE_D1_AIA(W, L,VRI,PRR)
# define _ARM_VLD1_XLANE_D1_AIA_64(W, L,VRI,PRR)    _ARM_VLD1_D1_AIA(W, VRI,PRR)
# define _ARM_VST1_XLANE_D1_AIA_8(W, L,VRI,PRR)     _ARM_VST1_LANE_D1_AIA(W, L,VRI,PRR)
# define _ARM_VST1_XLANE_D1_AIA_16(W, L,VRI,PRR)    _ARM_VST1_LANE_D1_AIA(W, L,VRI,PRR)
# define _ARM_VST1_XLANE_D1_AIA_32(W, L,VRI,PRR)    _ARM_VST1_LANE_D1_AIA(W, L,VRI,PRR)
# define _ARM_VST1_XLANE_D1_AIA_64(W, L,VRI,PRR)    _ARM_VST1_D1_AIA(W, VRI,PRR)

# define _ARM_VLD1_XLANE_D1_IA_8(W, L,VRI,PRR,ORR)  _ARM_VLD1_LANE_D1_AIA(W, L,VRI,PRR,ORR)
# define _ARM_VLD1_XLANE_D1_IA_16(W, L,VRI,PRR,ORR) _ARM_VLD1_LANE_D1_AIA(W, L,VRI,PRR,ORR)
# define _ARM_VLD1_XLANE_D1_IA_32(W, L,VRI,PRR,ORR) _ARM_VLD1_LANE_D1_AIA(W, L,VRI,PRR,ORR)
# define _ARM_VLD1_XLANE_D1_IA_64(W, L,VRI,PRR,ORR) _ARM_VLD1_D1_AIA(W, VRI,PRR,ORR)
# define _ARM_VST1_XLANE_D1_IA_8(W, L,VRI,PRR,ORR)  _ARM_VST1_LANE_D1_AIA(W, L,VRI,PRR,ORR)
# define _ARM_VST1_XLANE_D1_IA_16(W, L,VRI,PRR,ORR) _ARM_VST1_LANE_D1_AIA(W, L,VRI,PRR,ORR)
# define _ARM_VST1_XLANE_D1_IA_32(W, L,VRI,PRR,ORR) _ARM_VST1_LANE_D1_AIA(W, L,VRI,PRR,ORR)
# define _ARM_VST1_XLANE_D1_IA_64(W, L,VRI,PRR,ORR) _ARM_VST1_D1_AIA(W, VRI,PRR,ORR)
/* @} */

/* ========================================================================= */

/*
 * Store multiple N-element structures
 *
 * !!!NOTE!!!
 * Count of structures used is less then vector register size.
 * For example: only 5 of 8 1-element structures have to be stored.
 */

/*
 * 1) Store N 1-element structures to memory (PRR: base address) from vector register.
 * 2) Store N 1-element structures to memory (PRR: base address) from vector register, 0 ~ 8(16,32,64).
 *
 * Variant 1)
 * Vector size    Datatype     N
 * (D) 8           int-8     [1-7]
 * (D) 8           int-16    [1-3]
 * (D) 8           int-32    [1-1]
 * (Q) 16          int-8     [1-15]
 * (Q) 16          int-16    [1-7]
 * (Q) 16          int-32    [1-3]
 * (.) 32          int-8     [1-31]
 * (.) 32          int-16    [1-15]
 * (.) 32          int-32    [1-7]
 *
 * Variant 2)
 * Vector size    Datatype     N
 * (D) 8           int-8     [0-7]  --> [ 8,1-7]
 * (D) 8           int-16    [0-3]  --> [ 4,1-3]
 * (D) 8           int-32    [0-1]  --> [ 2,1-1]
 * (Q) 16          int-8     [0-15] --> [16,1-15]
 * (Q) 16          int-16    [0-7]  --> [ 8,1-7]
 * (Q) 16          int-32    [0-3]  --> [ 4,1-3]
 * (.) 32          int-8     [0-31] --> [32,1-31]
 * (.) 32          int-16    [0-15] --> [16,1-15]
 * (.) 32          int-32    [0-7]  --> [ 8,1-7]
 */

	/*
	 * 1-element structures
	 */

/* NRR in elements */
/* NRR changed */
# define _ARM_VST1_N_D1_AIA(W, NRR,VRI,PRR)             VIPM_GLUE(_ARM_VST1_N_D1_AIA_,W)(NRR,VRI,PRR)
# define _ARM_VST1_N_Q1_AIA(W, NRR,VRI,PRR)             VIPM_GLUE(_ARM_VST1_N_Q1_AIA_,W)(NRR,VRI,PRR)
# define _ARM_VST1_N_Q2_AIA(W, NRR,VRI1,VRI2,PRR)       VIPM_GLUE(_ARM_VST1_N_Q2_AIA_,W)(NRR,VRI1,VRI2,PRR)

/* NRR is not changed */
# define _ARM_VST1_NC0_D1_AIA(W, NRR,VRI,PRR,tmpreg)    VIPM_GLUE(_ARM_VST1_NC0_D1_AIA_,W)(tmpreg,NRR,VRI,PRR)
# define _ARM_VST1_NC0_Q1_AIA(W, NRR,VRI,PRR,tmpreg)    VIPM_GLUE(_ARM_VST1_NC0_Q1_AIA_,W)(tmpreg,NRR,VRI,PRR)
/* NRR changed */
# define _ARM_VST1_NC00_D1_AIA(W, NRR,VRI,PRR)          _ARM_VST1_NC0_D1_AIA(W, NRR,VRI,PRR,NRR)
# define _ARM_VST1_NC00_Q1_AIA(W, NRR,VRI,PRR)          _ARM_VST1_NC0_Q1_AIA(W, NRR,VRI,PRR,NRR)


/* NRR in bytes */
/* NRR is not changed */
# define _ARM_VST1_NC0B_D1_AIA(W, NRR,VRI,PRR,tmpreg)   VIPM_GLUE(_ARM_VST1_NC0B_D1_AIA_,W)(tmpreg,NRR,VRI,PRR)
# define _ARM_VST1_NC0B_Q1_AIA(W, NRR,VRI,PRR,tmpreg)   VIPM_GLUE(_ARM_VST1_NC0B_Q1_AIA_,W)(tmpreg,NRR,VRI,PRR)
# define _ARM_VST1_NC0B_Q2_AIA(W, NRR,VRI1,VRI2,PRR,tmpreg) \
	VIPM_GLUE(_ARM_VST1_NC0B_Q2_AIA_,W)(tmpreg,NRR,VRI1,VRI2,PRR)
# define _ARM_VST1_NC0B_Q4_AIA(W, NRR,VRI1,VRI2,VRI3,VRI4,PRR,tmpreg) \
	VIPM_GLUE(_ARM_VST1_NC0B_Q4_AIA_,W)(tmpreg,NRR,VRI1,VRI2,VRI3,VRI4,PRR)

/* ------------------------------------------------------------------------- */

	/*
	 * 2-element structures
	 */

/* NRR in elements */
/* NRR changed */
# define _ARM_VST2_N_D2_AIA(W, NRR,VRI0,VRI1,PRR)           VIPM_GLUE(_ARM_VST2_N_D2_AIA_,W)(NRR,VRI0,VRI1,PRR)
# define _ARM_VST2_N_Q2_AIA(W, NRR,VRI0,VRI1,PRR)           VIPM_GLUE(_ARM_VST2_N_Q2_AIA_,W)(NRR,VRI0,VRI1,PRR)

/* NRR is not changed */
# define _ARM_VST2_NC0_D2_AIA(W, NRR,VRI0,VRI1,PRR,tmpreg)  VIPM_GLUE(_ARM_VST2_NC0_D2_AIA_,W)(tmpreg,NRR,VRI0,VRI1,PRR)
# define _ARM_VST2_NC0_Q2_AIA(W, NRR,VRI0,VRI1,PRR,tmpreg)  VIPM_GLUE(_ARM_VST2_NC0_Q2_AIA_,W)(tmpreg,NRR,VRI0,VRI1,PRR)
/* NRR changed */
# define _ARM_VST2_NC00_D2_AIA(W, NRR,VRI0,VRI1,PRR)        _ARM_VST2_NC0_D2_AIA(W, NRR,VRI0,VRI1,PRR,NRR)
# define _ARM_VST2_NC00_Q2_AIA(W, NRR,VRI0,VRI1,PRR)        _ARM_VST2_NC0_Q2_AIA(W, NRR,VRI0,VRI1,PRR,NRR)


/* NRR in bytes */
/* NRR is not changed */
# define _ARM_VST2_NC0B_D2_AIA(W, NRR,VRI0,VRI1,PRR,tmpreg) VIPM_GLUE(_ARM_VST2_NC0B_D2_AIA_,W)(tmpreg,NRR,VRI0,VRI1,PRR)
# define _ARM_VST2_NC0B_Q2_AIA(W, NRR,VRI0,VRI1,PRR,tmpreg) VIPM_GLUE(_ARM_VST2_NC0B_Q2_AIA_,W)(tmpreg,NRR,VRI0,VRI1,PRR)

/* ------------------------------------------------------------------------- */

	/*
	 * 3-element structures
	 */

/* NRR in elements */
/* NRR changed */
# define _ARM_VST3_N_D3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR)           VIPM_GLUE(_ARM_VST3_N_D3_AIA_,W)(NRR,VRI0,VRI1,VRI2,PRR)
# define _ARM_VST3_N_Q3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR)           VIPM_GLUE(_ARM_VST3_N_Q3_AIA_,W)(NRR,VRI0,VRI1,VRI2,PRR)

/* NRR is not changed */
# define _ARM_VST3_NC0_D3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR,tmpreg)  VIPM_GLUE(_ARM_VST3_NC0_D3_AIA_,W)(tmpreg,NRR,VRI0,VRI1,VRI2,PRR)
# define _ARM_VST3_NC0_Q3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR,tmpreg)  VIPM_GLUE(_ARM_VST3_NC0_Q3_AIA_,W)(tmpreg,NRR,VRI0,VRI1,VRI2,PRR)
/* NRR changed */
# define _ARM_VST3_NC00_D3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR)        _ARM_VST3_NC0_D3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR,NRR)
# define _ARM_VST3_NC00_Q3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR)        _ARM_VST3_NC0_Q3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR,NRR)


/* NRR in bytes */
/* NRR is not changed */
# define _ARM_VST3_NC0B_D3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR,tmpreg) VIPM_GLUE(_ARM_VST3_NC0B_D3_AIA_,W)(tmpreg,NRR,VRI0,VRI1,VRI2,PRR)
# define _ARM_VST3_NC0B_Q3_AIA(W, NRR,VRI0,VRI1,VRI2,PRR,tmpreg) VIPM_GLUE(_ARM_VST3_NC0B_Q3_AIA_,W)(tmpreg,NRR,VRI0,VRI1,VRI2,PRR)

/* ------------------------------------------------------------------------- */

	/*
	 * 4-element structures
	 */

/* NRR in elements */
/* NRR changed */
# define _ARM_VST4_N_D4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR)             VIPM_GLUE(_ARM_VST4_N_D4_AIA_,W)(NRR,VRI0,VRI1,VRI2,VRI3,PRR)
# define _ARM_VST4_N_Q4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR)             VIPM_GLUE(_ARM_VST4_N_Q4_AIA_,W)(NRR,VRI0,VRI1,VRI2,VRI3,PRR)

/* NRR is not changed */
# define _ARM_VST4_NC0_D4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR,tmpreg)    VIPM_GLUE(_ARM_VST4_NC0_D4_AIA_,W)(tmpreg,NRR,VRI0,VRI1,VRI2,VRI3,PRR)
# define _ARM_VST4_NC0_Q4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR,tmpreg)    VIPM_GLUE(_ARM_VST4_NC0_Q4_AIA_,W)(tmpreg,NRR,VRI0,VRI1,VRI2,VRI3,PRR)
/* NRR changed */
# define _ARM_VST4_NC00_D4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR)          _ARM_VST4_NC0_D4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR,NRR)
# define _ARM_VST4_NC00_Q4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR)          _ARM_VST4_NC0_Q4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR,NRR)


/* NRR in bytes */
/* NRR is not changed */
# define _ARM_VST4_NC0B_D4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR,tmpreg) \
    VIPM_GLUE(_ARM_VST4_NC0B_D4_AIA_,W)(tmpreg,NRR,VRI0,VRI1,VRI2,VRI3,PRR)
# define _ARM_VST4_NC0B_Q4_AIA(W, NRR,VRI0,VRI1,VRI2,VRI3,PRR,tmpreg) \
    VIPM_GLUE(_ARM_VST4_NC0B_Q4_AIA_,W)(tmpreg,NRR,VRI0,VRI1,VRI2,VRI3,PRR)

/* ------------------------------------------------------------------------- */

/* impl.
 * @{
 */

	/*
	 * 1-element structures
	 */

# define _ARM_VST1_N_D1_AIA_8(NRR,VRI,PRR)      \
    _ARM_IF_CMPI(GE, NRR, 4, 11004f)            \
        _ARM_VST1_LANE_D1_AIA(32, 0,VRI, PRR)   \
        _ARM_SUBI(NRR, NRR, 4)                  \
        _ARM_VEROTL_D(32, VRI,VRI,1)            \
    _ARM_ENDIF(11004)                           \
    _ARM_IF_CMPI(GE, NRR, 2, 11002f)            \
        _ARM_VST1_LANE_D1_AIA(16, 0,VRI, PRR)   \
        _ARM_SUBI(NRR, NRR, 2)                  \
        _ARM_VEROTL_D(16, VRI,VRI,1)            \
    _ARM_ENDIF(11002)                           \
    _ARM_IF_CMPI(GE, NRR, 1, 11001f)            \
        _ARM_VST1_LANE_D1_AIA(8, 0,VRI, PRR)    \
    _ARM_ENDIF(11001)

# define _ARM_VST1_N_D1_AIA_16(NRR,VRI,PRR)     \
    _ARM_IF_CMPI(GE, NRR, 2, 11002f)            \
       _ARM_VST1_LANE_D1_AIA(32, 0,VRI, PRR)    \
       _ARM_SUBI(NRR, NRR, 2)                   \
       _ARM_VEROTL_D(32, VRI,VRI,1)             \
    _ARM_ENDIF(11002)                           \
    _ARM_IF_CMPI(GE, NRR, 1, 11001f)            \
       _ARM_VST1_LANE_D1_AIA(16, 0,VRI, PRR)    \
    _ARM_ENDIF(11001)

# define _ARM_VST1_N_D1_AIA_32(NRR,VRI,PRR)     \
    _ARM_IF_CMPI(GE, NRR, 1, 11001f)            \
       _ARM_VST1_LANE_D1_AIA(32, 0,VRI, PRR)    \
    _ARM_ENDIF(11001)


# define _ARM_VST1_N_Q1_AIA_8(NRR,VRI,PRR)              \
    _ARM_IF_CMPI(GE, NRR, 8, 11008f)                    \
        _ARM_VST1_D1_AIA(8, _ARM_QDMAP0(VRI), PRR)      \
        _ARM_SUBI(NRR, NRR, 8)                          \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI),_ARM_QDMAP1(VRI))  \
    _ARM_ENDIF(11008)                                   \
	_ARM_VST1_N_D1_AIA_8(NRR,_ARM_QDMAP0(VRI),PRR)

# define _ARM_VST1_N_Q1_AIA_16(NRR,VRI,PRR)             \
    _ARM_IF_CMPI(GE, NRR, 4, 11004f)                    \
        _ARM_VST1_D1_AIA(16, _ARM_QDMAP0(VRI),PRR)      \
        _ARM_SUBI(NRR, NRR, 4)                          \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI),_ARM_QDMAP1(VRI))  \
    _ARM_ENDIF(11004)                                   \
	_ARM_VST1_N_D1_AIA_16(NRR,_ARM_QDMAP0(VRI),PRR)

# define _ARM_VST1_N_Q1_AIA_32(NRR,VRI,PRR)             \
    _ARM_IF_CMPI(GE, NRR, 2, 11002f)                    \
        _ARM_VST1_D1_AIA(32, _ARM_QDMAP0(VRI),PRR)      \
        _ARM_SUBI(NRR, NRR, 2)                          \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI),_ARM_QDMAP1(VRI))  \
    _ARM_ENDIF(11002)                                   \
	_ARM_VST1_N_D1_AIA_32(NRR,_ARM_QDMAP0(VRI),PRR)


# define _ARM_VST1_N_Q2_AIA_8(NRR,VRI1,VRI2,PRR)        \
    _ARM_IF_CMPI(GE, NRR, 16, 11008f)                   \
        _ARM_VST1_Q1_AIA(8, VRI1, PRR)                  \
        _ARM_SUBI(NRR, NRR, 16)                         \
        _ARM_VMOV_Q(VRI1,VRI2)                          \
    _ARM_ENDIF(11008)                                   \
	_ARM_VST1_N_Q1_AIA_8(NRR,VRI1,PRR)

# define _ARM_VST1_N_Q2_AIA_16(NRR,VRI1,VRI2,PRR)       \
    _ARM_IF_CMPI(GE, NRR, 8, 11004f)                    \
        _ARM_VST1_Q1_AIA(16, VRI1, PRR)                 \
        _ARM_SUBI(NRR, NRR, 8)                          \
        _ARM_VMOV_Q(VRI1,VRI2)                          \
    _ARM_ENDIF(11004)                                   \
	_ARM_VST1_N_Q1_AIA_16(NRR,VRI1,PRR)

# define _ARM_VST1_N_Q2_AIA_32(NRR,VRI1,VRI2,PRR)       \
    _ARM_IF_CMPI(GE, NRR, 4, 11002f)                    \
        _ARM_VST1_Q1_AIA(32, VRI1, PRR)                 \
        _ARM_SUBI(NRR, NRR, 4)                          \
        _ARM_VMOV_Q(VRI1,VRI2)                          \
    _ARM_ENDIF(11002)                                   \
	_ARM_VST1_N_Q1_AIA_32(NRR,VRI1,PRR)



# define _ARM_VST1_NC0_D1_AIA_8(TRR,NRR,VRI,PRR)    \
	_ARM_IF_ANDI_NE0(TRR,NRR,7, 11000f)             \
    	_ARM_VST1_N_D1_AIA_8(TRR,VRI,PRR)           \
	    _ARM_GOTO_IMM(11007f)                       \
	_ARM_ENDIF(11000)                               \
    _ARM_VST1_D1_AIA(8, VRI, PRR)                   \
	"11007: \n\t"

# define _ARM_VST1_NC0_D1_AIA_16(TRR,NRR,VRI,PRR)   \
	_ARM_IF_ANDI_NE0(TRR,NRR,3, 11000f)             \
    	_ARM_VST1_N_D1_AIA_16(TRR,VRI,PRR)          \
	    _ARM_GOTO_IMM(11007f)                       \
	_ARM_ENDIF(11000)                               \
    _ARM_VST1_D1_AIA(16, VRI, PRR)                  \
	"11007: \n\t"

# define _ARM_VST1_NC0_D1_AIA_32(TRR,NRR,VRI,PRR)   \
	_ARM_IF_ANDI_NE0(TRR,NRR,1, 11001f)             \
    	_ARM_VST1_LANE_D1_AIA(32, 0,VRI, PRR)       \
	    _ARM_GOTO_IMM(11007f)                       \
	_ARM_ENDIF(11001)                               \
	_ARM_VST1_D1_AIA(32, VRI, PRR)                  \
	"11007: \n\t"


# define _ARM_VST1_NC0_Q1_AIA_8(TRR,NRR,VRI,PRR)    \
	_ARM_IF_ANDI_NE0(TRR,NRR,15, 11000f)            \
    	_ARM_VST1_N_Q1_AIA_8(TRR,VRI,PRR)           \
	    _ARM_GOTO_IMM(11015f)                       \
	_ARM_ENDIF(11000)                               \
    _ARM_VST1_Q1_AIA(8, VRI, PRR)                   \
	"11015: \n\t"

# define _ARM_VST1_NC0_Q1_AIA_16(TRR,NRR,VRI,PRR)   \
	_ARM_IF_ANDI_NE0(TRR,NRR,7, 11000f)             \
    	_ARM_VST1_N_Q1_AIA_16(TRR,VRI,PRR)          \
	    _ARM_GOTO_IMM(11015f)                       \
	_ARM_ENDIF(11000)                               \
    _ARM_VST1_Q1_AIA(16, VRI, PRR)                  \
	"11015: \n\t"

# define _ARM_VST1_NC0_Q1_AIA_32(TRR,NRR,VRI,PRR)   \
	_ARM_IF_ANDI_NE0(TRR,NRR,3, 11000f)             \
	    _ARM_VST1_N_Q1_AIA_32(TRR,VRI,PRR)          \
	    _ARM_GOTO_IMM(11015f)                       \
	_ARM_ENDIF(11000)                               \
    _ARM_VST1_Q1_AIA(32, VRI, PRR)                  \
	"11015: \n\t"



# define _ARM_VST1_NC0B_D1_AIA_8(TRR,NRR,VRI,PRR)   \
	_ARM_VST1_NC0_D1_AIA_8(TRR,NRR,VRI,PRR)

# define _ARM_VST1_NC0B_D1_AIA_16(TRR,NRR,VRI,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,1)                         \
	_ARM_VST1_NC0_D1_AIA_16(TRR,TRR,VRI,PRR)

# define _ARM_VST1_NC0B_D1_AIA_32(TRR,NRR,VRI,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,2)                         \
	_ARM_VST1_NC0_D1_AIA_32(TRR,TRR,VRI,PRR)


# define _ARM_VST1_NC0B_Q1_AIA_8(TRR,NRR,VRI,PRR)   \
	_ARM_VST1_NC0_Q1_AIA_8(TRR,NRR,VRI,PRR)

# define _ARM_VST1_NC0B_Q1_AIA_16(TRR,NRR,VRI,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,1)                         \
	_ARM_VST1_NC0_Q1_AIA_16(TRR,TRR,VRI,PRR)

# define _ARM_VST1_NC0B_Q1_AIA_32(TRR,NRR,VRI,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,2)                         \
	_ARM_VST1_NC0_Q1_AIA_32(TRR,TRR,VRI,PRR)


# define _ARM_VST1_NC0B_Q2_AIA_8(TRR,NRR,VRI1,VRI2,PRR)  \
	_ARM_ANDI(TRR,TRR,31)                                \
	_ARM_IF_EQ0(TRR, 11000f)                             \
		_ARM_VST1_Q2_AIA(8, VRI1,VRI2,PRR)               \
		_ARM_GOTO_IMM(11100f)                            \
	_ARM_ENDIF(11000)                                    \
	_ARM_IF_GEI(TRR, 16, 11016f)                         \
		_ARM_VST1_Q1_AIA(8, VRI1, PRR)                   \
		_ARM_SUBI(TRR,TRR, 16)                           \
		_ARM_VST1_N_Q1_AIA_8(TRR, VRI2, PRR)             \
		_ARM_GOTO_IMM(11100f)                            \
	_ARM_ENDIF(11016)                                    \
	_ARM_IF_LTI(TRR, 16, 11014f)                         \
		_ARM_VST1_N_Q1_AIA_8(TRR,VRI1,PRR)               \
		_ARM_GOTO_IMM(11100f)                            \
	_ARM_ENDIF(11014)                                    \
	"11100:"

# define _ARM_VST1_NC0B_Q2_AIA_16(TRR,NRR,VRI1,VRI2,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,1)                               \
	_ARM_ANDI(TRR,TRR,15)                                 \
	_ARM_IF_EQ0(TRR, 11000f)                              \
		_ARM_VST1_Q2_AIA(16, VRI1,VRI2,PRR)               \
		_ARM_GOTO_IMM(11100f)                             \
	_ARM_ENDIF(11000)                                     \
	_ARM_IF_GEI(TRR, 8, 11016f)                           \
		_ARM_VST1_Q1_AIA(16, VRI1, PRR)                   \
		_ARM_SUBI(TRR,TRR, 8)                             \
		_ARM_VST1_N_Q1_AIA_16(TRR, VRI2, PRR)             \
		_ARM_GOTO_IMM(11100f)                             \
	_ARM_ENDIF(11016)                                     \
	_ARM_IF_LTI(TRR, 8, 11014f)                           \
		_ARM_VST1_N_Q1_AIA_16(TRR,VRI1,PRR)               \
		_ARM_GOTO_IMM(11100f)                             \
	_ARM_ENDIF(11014)                                     \
	"11100:"

# define _ARM_VST1_NC0B_Q2_AIA_32(TRR,NRR,VRI1,VRI2,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,2)                               \
	_ARM_ANDI(TRR,TRR,7)                                  \
	_ARM_IF_EQ0(TRR, 11000f)                              \
		_ARM_VST1_Q2_AIA(32, VRI1,VRI2,PRR)               \
		_ARM_GOTO_IMM(11100f)                             \
	_ARM_ENDIF(11000)                                     \
	_ARM_IF_GEI(TRR, 4, 11032f)                           \
		_ARM_VST1_Q1_AIA(32, VRI1, PRR)                   \
		_ARM_SUBI(TRR,TRR, 4)                             \
		_ARM_VST1_N_Q1_AIA_32(TRR, VRI2, PRR)             \
		_ARM_GOTO_IMM(11100f)                             \
	_ARM_ENDIF(11032)                                     \
	_ARM_IF_LTI(TRR, 4, 11031f)                           \
		_ARM_VST1_N_Q1_AIA_32(TRR,VRI1,PRR)               \
		_ARM_GOTO_IMM(11100f)                             \
	_ARM_ENDIF(11031)                                     \
	"11100:"


# define _ARM_VST1_NC0B_Q4_AIA_32(TRR,NRR,VRI1,VRI2,VRI3,VRI4,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,2)                                         \
	_ARM_ANDI(TRR,TRR,15)                                           \
	_ARM_IF_EQ0(TRR, 11000f)                                        \
		_ARM_VST1_Q4_AIA(32, VRI1,VRI2,VRI3,VRI4,PRR)               \
		_ARM_GOTO_IMM(11100f)                                       \
	_ARM_ENDIF(11000)                                               \
	_ARM_IF_GEI(TRR, 12, 11048f)                                    \
		_ARM_SUBI(TRR,TRR, 12)                                      \
		_ARM_VST1_Q2_AIA(32, VRI1,VRI2,PRR)                         \
		_ARM_VST1_Q1_AIA(32, VRI3,PRR)                              \
		_ARM_VST1_N_Q1_AIA_32(TRR,VRI4,PRR)                         \
		_ARM_GOTO_IMM(11100f)                                       \
	_ARM_ENDIF(11048)                                               \
	_ARM_IF_GEI(TRR, 8, 11032f)                                     \
		_ARM_SUBI(TRR,TRR, 8)                                       \
		_ARM_VST1_Q2_AIA(32, VRI1,VRI2,PRR)                         \
		_ARM_VST1_N_Q1_AIA_32(TRR,VRI3,PRR)                         \
		_ARM_GOTO_IMM(11100f)                                       \
	_ARM_ENDIF(11032)                                               \
	_ARM_IF_GEI(TRR, 4, 11016f)                                     \
		_ARM_SUBI(TRR,TRR, 4)                                       \
		_ARM_VST1_Q1_AIA(32, VRI1,PRR)                              \
		_ARM_VST1_N_Q1_AIA_32(TRR,VRI2,PRR)                         \
		_ARM_GOTO_IMM(11100f)                                       \
	_ARM_ENDIF(11016)                                               \
	_ARM_IF_LTI(TRR, 4, 11014f)                                     \
		_ARM_VST1_N_Q1_AIA_32(TRR,VRI1,PRR)                         \
	_ARM_ENDIF(11014)                                               \
	"11100:"


	/*
	 * 3-element structures
	 */

# define _ARM_VST3_N_D3_AIA_8(NRR,VRI0,VRI1,VRI2,PRR)     \
    _ARM_IF_CMPI(GE, NRR, 4, 11004f)                      \
        _ARM_VST3_LANE_D3_AIA(8, 0,VRI0,VRI1,VRI2,PRR)    \
        _ARM_VST3_LANE_D3_AIA(8, 1,VRI0,VRI1,VRI2,PRR)    \
        _ARM_VST3_LANE_D3_AIA(8, 2,VRI0,VRI1,VRI2,PRR)    \
        _ARM_VST3_LANE_D3_AIA(8, 3,VRI0,VRI1,VRI2,PRR)    \
        _ARM_SUBI(NRR, NRR, 4)                            \
        _ARM_VEROTL_D(32, VRI0,VRI0,1)                    \
        _ARM_VEROTL_D(32, VRI1,VRI1,1)                    \
        _ARM_VEROTL_D(32, VRI2,VRI2,1)                    \
    _ARM_ENDIF(11004)                                     \
    _ARM_IF_CMPI(GE, NRR, 2, 11002f)                      \
        _ARM_VST3_LANE_D3_AIA(8, 0,VRI0,VRI1,VRI2,PRR)    \
        _ARM_VST3_LANE_D3_AIA(8, 1,VRI0,VRI1,VRI2,PRR)    \
        _ARM_SUBI(NRR, NRR, 2)                            \
        _ARM_VEROTL_D(16, VRI0,VRI0,1)                    \
        _ARM_VEROTL_D(16, VRI1,VRI1,1)                    \
        _ARM_VEROTL_D(16, VRI2,VRI2,1)                    \
    _ARM_ENDIF(11002)                                     \
    _ARM_IF_CMPI(GE, NRR, 1, 11001f)                      \
        _ARM_VST3_LANE_D3_AIA(8, 0,VRI0,VRI1,VRI2,PRR)    \
    _ARM_ENDIF(11001)

# define _ARM_VST3_N_D3_AIA_16(NRR,VRI0,VRI1,VRI2,PRR)    \
    _ARM_IF_CMPI(GE, NRR, 2, 11002f)                      \
        _ARM_VST3_LANE_D3_AIA(16, 0,VRI0,VRI1,VRI2,PRR)   \
        _ARM_VST3_LANE_D3_AIA(16, 1,VRI0,VRI1,VRI2,PRR)   \
        _ARM_SUBI(NRR, NRR, 2)                            \
        _ARM_VEROTL_D(32, VRI0,VRI0,1)                    \
        _ARM_VEROTL_D(32, VRI1,VRI1,1)                    \
        _ARM_VEROTL_D(32, VRI2,VRI2,1)                    \
    _ARM_ENDIF(11002)                                     \
    _ARM_IF_CMPI(GE, NRR, 1, 11001f)                      \
        _ARM_VST3_LANE_D3_AIA(16, 0,VRI0,VRI1,VRI2,PRR)   \
    _ARM_ENDIF(11001)

# define _ARM_VST3_N_D3_AIA_32(NRR,VRI0,VRI1,VRI2,PRR)    \
    _ARM_IF_CMPI(GE, NRR, 1, 11001f)                      \
        _ARM_VST3_LANE_D3_AIA(32, 0,VRI0,VRI1,VRI2,PRR)   \
    _ARM_ENDIF(11001)


# define _ARM_VST3_N_Q3_AIA_8(NRR,VRI0,VRI1,VRI2,PRR)     \
    _ARM_IF_CMPI(GE, NRR, 8, 11008f)                      \
        _ARM_VST3_D3_AIA(8, _ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),PRR) \
        _ARM_SUBI(NRR, NRR, 8)                            \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI0),_ARM_QDMAP1(VRI0))  \
        _ARM_VMOV_D(_ARM_QDMAP1(VRI0),_ARM_QDMAP1(VRI1))  \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI1),_ARM_QDMAP1(VRI2))  \
		_ARM_GOTO_IMM(11007f)                             \
    _ARM_ENDIF(11008)                                     \
		_ARM_VMOV_D(_ARM_QDMAP1(VRI0),_ARM_QDMAP0(VRI1))  \
		_ARM_VMOV_D(_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2))  \
	"11007: \n\t"                                         \
	_ARM_VST3_N_D3_AIA_8(NRR,_ARM_QDMAP0(VRI0),_ARM_QDMAP1(VRI0),_ARM_QDMAP0(VRI1),PRR)

# define _ARM_VST3_N_Q3_AIA_16(NRR,VRI0,VRI1,VRI2,PRR)    \
    _ARM_IF_CMPI(GE, NRR, 4, 11004f)                      \
        _ARM_VST3_D3_AIA(16, _ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),PRR) \
        _ARM_SUBI(NRR, NRR, 4)                            \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI0),_ARM_QDMAP1(VRI0))  \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI1),_ARM_QDMAP1(VRI1))  \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI2),_ARM_QDMAP1(VRI2))  \
    _ARM_ENDIF(11004)                                     \
	_ARM_VST3_N_D3_AIA_16(NRR,_ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),PRR)

# define _ARM_VST3_N_Q3_AIA_32(NRR,VRI0,VRI1,VRI2,PRR)    \
    _ARM_IF_CMPI(GE, NRR, 2, 11002f)                      \
        _ARM_VST3_D3_AIA(32, _ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),PRR) \
        _ARM_SUBI(NRR, NRR, 2)                            \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI0),_ARM_QDMAP1(VRI0))  \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI1),_ARM_QDMAP1(VRI1))  \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI2),_ARM_QDMAP1(VRI2))  \
    _ARM_ENDIF(11002)                                     \
	_ARM_VST3_N_D3_AIA_32(NRR,_ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),PRR)



# define _ARM_VST3_NC0_D3_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,PRR)    \
_ARM_IF_ANDI_NE0(TRR,NRR,7, 11000f)                            \
    _ARM_VST3_N_D3_AIA_8(TRR,VRI0,VRI1,VRI2,PRR)               \
    _ARM_GOTO_IMM(11007f)                                      \
_ARM_ENDIF(11000)                                              \
    _ARM_VST3_D3_AIA(8, VRI0,VRI1,VRI2,PRR)                    \
"11007: \n\t"

# define _ARM_VST3_NC0_D3_AIA_16(TRR,NRR,VRI0,VRI1,VRI2,PRR)   \
_ARM_IF_ANDI_NE0(TRR,NRR,3, 11000f)                            \
    _ARM_VST3_N_D3_AIA_16(TRR,VRI0,VRI1,VRI2,PRR)              \
    _ARM_GOTO_IMM(11007f)                                      \
_ARM_ENDIF(11000)                                              \
    _ARM_VST3_D3_AIA(16, VRI0,VRI1,VRI2,PRR)                   \
"11007: \n\t"

# define _ARM_VST3_NC0_D3_AIA_32(TRR,NRR,VRI0,VRI1,VRI2,PRR)   \
_ARM_IF_ANDI_NE0(TRR,NRR,1, 11001f)                            \
    _ARM_VST3_LANE_D3_AIA(32, 0,VRI0,VRI1,VRI2,PRR)            \
    _ARM_GOTO_IMM(11007f)                                      \
_ARM_ENDIF(11001)                                              \
    _ARM_VST3_D3_AIA(32, VRI0,VRI1,VRI2,PRR)                   \
"11007: \n\t"


# define _ARM_VST3_NC0_Q3_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,PRR)    \
_ARM_IF_ANDI_NE0(TRR,NRR,15, 11000f)                           \
    _ARM_VST3_N_Q3_AIA_8(TRR,VRI0,VRI1,VRI2,PRR)               \
    _ARM_GOTO_IMM(11015f)                                      \
_ARM_ENDIF(11000)                                              \
    _ARM_VST3_Q3_AIA(8, VRI0,VRI1,VRI2,PRR)                    \
"11015: \n\t"

# define _ARM_VST3_NC0_Q3_AIA_16(TRR,NRR,VRI0,VRI1,VRI2,PRR)   \
_ARM_IF_ANDI_NE0(TRR,NRR,7, 11000f)                            \
    _ARM_VST3_N_Q3_AIA_16(TRR,VRI0,VRI1,VRI2,PRR)              \
    _ARM_GOTO_IMM(11015f)                                      \
_ARM_ENDIF(11000)                                              \
    _ARM_VST3_Q3_AIA(16, VRI0,VRI1,VRI2,PRR)                   \
"11015: \n\t"

# define _ARM_VST3_NC0_Q3_AIA_32(TRR,NRR,VRI0,VRI1,VRI2,PRR)   \
_ARM_IF_ANDI_NE0(TRR,NRR,3, 11000f)                            \
    _ARM_VST3_N_Q3_AIA_32(TRR,VRI0,VRI1,VRI2,PRR)              \
    _ARM_GOTO_IMM(11015f)                                      \
_ARM_ENDIF(11000)                                              \
    _ARM_VST3_Q3_AIA(32, VRI0,VRI1,VRI2,PRR)                   \
"11015: \n\t"



# define _ARM_VST3_NC0B_D3_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,PRR)   \
	_ARM_VST3_NC0_D3_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,PRR)

# define _ARM_VST3_NC0B_D3_AIA_16(TRR,NRR,VRI0,VRI1,VRI2,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,1)                                    \
	_ARM_VST3_NC0_D3_AIA_16(TRR,TRR,VRI0,VRI1,VRI2,PRR)

# define _ARM_VST3_NC0B_D3_AIA_32(TRR,NRR,VRI0,VRI1,VRI2,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,2)                                    \
	_ARM_VST3_NC0_D3_AIA_32(TRR,TRR,VRI0,VRI1,VRI2,PRR)


# define _ARM_VST3_NC0B_Q3_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,PRR)   \
	_ARM_VST3_NC0_Q3_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,PRR)

# define _ARM_VST3_NC0B_Q3_AIA_16(TRR,NRR,VRI0,VRI1,VRI2,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,1)                                    \
	_ARM_VST3_NC0_Q3_AIA_16(TRR,TRR,VRI0,VRI1,VRI2,PRR)

# define _ARM_VST3_NC0B_Q3_AIA_32(TRR,NRR,VRI0,VRI1,VRI2,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,2)                                    \
	_ARM_VST3_NC0_Q3_AIA_32(TRR,TRR,VRI0,VRI1,VRI2,PRR)


	/*
	 * 4-element structures
	 */

# define _ARM_VST4_N_D4_AIA_8(NRR,VRI0,VRI1,VRI2,VRI3,PRR)     \
    _ARM_IF_CMPI(GE, NRR, 4, 11004f)                           \
        _ARM_VST4_LANE_D4_AIA(8, 0,VRI0,VRI1,VRI2,VRI3,PRR)    \
        _ARM_VST4_LANE_D4_AIA(8, 1,VRI0,VRI1,VRI2,VRI3,PRR)    \
        _ARM_VST4_LANE_D4_AIA(8, 2,VRI0,VRI1,VRI2,VRI3,PRR)    \
        _ARM_VST4_LANE_D4_AIA(8, 3,VRI0,VRI1,VRI2,VRI3,PRR)    \
        _ARM_SUBI(NRR, NRR, 4)                                 \
        _ARM_VEROTL_D(32, VRI0,VRI0,1)                         \
        _ARM_VEROTL_D(32, VRI1,VRI1,1)                         \
        _ARM_VEROTL_D(32, VRI2,VRI2,1)                         \
        _ARM_VEROTL_D(32, VRI3,VRI3,1)                         \
    _ARM_ENDIF(11004)                                          \
    _ARM_IF_CMPI(GE, NRR, 2, 11002f)                           \
        _ARM_VST4_LANE_D4_AIA(8, 0,VRI0,VRI1,VRI2,VRI3,PRR)    \
        _ARM_VST4_LANE_D4_AIA(8, 1,VRI0,VRI1,VRI2,VRI3,PRR)    \
        _ARM_SUBI(NRR, NRR, 2)                                 \
        _ARM_VEROTL_D(16, VRI0,VRI0,1)                         \
        _ARM_VEROTL_D(16, VRI1,VRI1,1)                         \
        _ARM_VEROTL_D(16, VRI2,VRI2,1)                         \
        _ARM_VEROTL_D(16, VRI3,VRI3,1)                         \
    _ARM_ENDIF(11002)                                          \
    _ARM_IF_CMPI(GE, NRR, 1, 11001f)                           \
        _ARM_VST4_LANE_D4_AIA(8, 0,VRI0,VRI1,VRI2,VRI3,PRR)    \
    _ARM_ENDIF(11001)

# define _ARM_VST4_N_D4_AIA_16(NRR,VRI0,VRI1,VRI2,VRI3,PRR)    \
    _ARM_IF_CMPI(GE, NRR, 2, 11002f)                           \
        _ARM_VST4_LANE_D4_AIA(16, 0,VRI0,VRI1,VRI2,VRI3,PRR)   \
        _ARM_VST4_LANE_D4_AIA(16, 1,VRI0,VRI1,VRI2,VRI3,PRR)   \
        _ARM_SUBI(NRR, NRR, 2)                                 \
        _ARM_VEROTL_D(32, VRI0,VRI0,1)                         \
        _ARM_VEROTL_D(32, VRI1,VRI1,1)                         \
        _ARM_VEROTL_D(32, VRI2,VRI2,1)                         \
        _ARM_VEROTL_D(32, VRI3,VRI3,1)                         \
    _ARM_ENDIF(11002)                                          \
    _ARM_IF_CMPI(GE, NRR, 1, 11001f)                           \
        _ARM_VST4_LANE_D4_AIA(16, 0,VRI0,VRI1,VRI2,VRI3,PRR)   \
    _ARM_ENDIF(11001)

# define _ARM_VST4_N_D4_AIA_32(NRR,VRI0,VRI1,VRI2,VRI3,PRR)    \
    _ARM_IF_CMPI(GE, NRR, 1, 11001f)                           \
        _ARM_VST4_LANE_D4_AIA(32, 0,VRI0,VRI1,VRI2,VRI3,PRR)   \
    _ARM_ENDIF(11001)


# define _ARM_VST4_N_Q4_AIA_8(NRR,VRI0,VRI1,VRI2,VRI3,PRR)     \
    _ARM_IF_CMPI(GE, NRR, 8, 11008f)                           \
        _ARM_VST4_D4_AIA(8, _ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),_ARM_QDMAP0(VRI3),PRR) \
        _ARM_SUBI(NRR, NRR, 8)                                 \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI0),_ARM_QDMAP1(VRI0))       \
        _ARM_VMOV_D(_ARM_QDMAP1(VRI0),_ARM_QDMAP1(VRI1))       \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI1),_ARM_QDMAP1(VRI2))       \
        _ARM_VMOV_D(_ARM_QDMAP1(VRI1),_ARM_QDMAP1(VRI3))       \
		_ARM_GOTO_IMM(11007f)                                  \
    _ARM_ENDIF(11008)                                          \
		_ARM_VMOV_D(_ARM_QDMAP1(VRI0),_ARM_QDMAP0(VRI1))       \
		_ARM_VMOV_D(_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2))       \
		_ARM_VMOV_D(_ARM_QDMAP1(VRI1),_ARM_QDMAP0(VRI3))       \
	"11007: \n\t"                                              \
	_ARM_VST4_N_D4_AIA_8(NRR,_ARM_QDMAP0(VRI0),_ARM_QDMAP1(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP1(VRI1),PRR)

# define _ARM_VST4_N_Q4_AIA_16(NRR,VRI0,VRI1,VRI2,VRI3,PRR)    \
    _ARM_IF_CMPI(GE, NRR, 4, 11004f)                           \
        _ARM_VST4_D4_AIA(16, _ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),_ARM_QDMAP0(VRI3),PRR) \
        _ARM_SUBI(NRR, NRR, 4)                                 \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI0),_ARM_QDMAP1(VRI0))       \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI1),_ARM_QDMAP1(VRI1))       \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI2),_ARM_QDMAP1(VRI2))       \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI3),_ARM_QDMAP1(VRI3))       \
    _ARM_ENDIF(11004)                                          \
	_ARM_VST4_N_D4_AIA_16(NRR,_ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),_ARM_QDMAP0(VRI3),PRR)

# define _ARM_VST4_N_Q4_AIA_32(NRR,VRI0,VRI1,VRI2,VRI3,PRR)    \
    _ARM_IF_CMPI(GE, NRR, 2, 11002f)                           \
        _ARM_VST4_D4_AIA(32, _ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),_ARM_QDMAP0(VRI3),PRR) \
        _ARM_SUBI(NRR, NRR, 2)                                 \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI0),_ARM_QDMAP1(VRI0))       \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI1),_ARM_QDMAP1(VRI1))       \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI2),_ARM_QDMAP1(VRI2))       \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI3),_ARM_QDMAP1(VRI3))       \
    _ARM_ENDIF(11002)                                          \
	_ARM_VST4_N_D4_AIA_32(NRR,_ARM_QDMAP0(VRI0),_ARM_QDMAP0(VRI1),_ARM_QDMAP0(VRI2),_ARM_QDMAP0(VRI3),PRR)



# define _ARM_VST4_NC0_D4_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)    \
_ARM_IF_ANDI_NE0(TRR,NRR,7, 11000f)                                 \
    _ARM_VST4_N_D4_AIA_8(TRR,VRI0,VRI1,VRI2,VRI3,PRR)               \
    _ARM_GOTO_IMM(11007f)                                           \
_ARM_ENDIF(11000)                                                   \
    _ARM_VST4_D4_AIA(8, VRI0,VRI1,VRI2,VRI3,PRR)                    \
"11007: \n\t"

# define _ARM_VST4_NC0_D4_AIA_16(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)   \
_ARM_IF_ANDI_NE0(TRR,NRR,3, 11000f)                                 \
    _ARM_VST4_N_D4_AIA_16(TRR,VRI0,VRI1,VRI2,VRI3,PRR)              \
    _ARM_GOTO_IMM(11007f)                                           \
_ARM_ENDIF(11000)                                                   \
    _ARM_VST4_D4_AIA(16, VRI0,VRI1,VRI2,VRI3,PRR)                   \
"11007: \n\t"

# define _ARM_VST4_NC0_D4_AIA_32(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)   \
_ARM_IF_ANDI_NE0(TRR,NRR,1, 11001f)                                 \
    _ARM_VST4_LANE_D4_AIA(32, 0,VRI0,VRI1,VRI2,VRI3,PRR)            \
    _ARM_GOTO_IMM(11007f)                                           \
_ARM_ENDIF(11001)                                                   \
    _ARM_VST4_D4_AIA(32, VRI0,VRI1,VRI2,VRI3,PRR)                   \
"11007: \n\t"


# define _ARM_VST4_NC0_Q4_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)    \
_ARM_IF_ANDI_NE0(TRR,NRR,15, 11000f)                                \
    _ARM_VST4_N_Q4_AIA_8(TRR,VRI0,VRI1,VRI2,VRI3,PRR)               \
    _ARM_GOTO_IMM(11015f)                                           \
_ARM_ENDIF(11000)                                                   \
    _ARM_VST4_Q4_AIA(8, VRI0,VRI1,VRI2,VRI3,PRR)                    \
"11015: \n\t"

# define _ARM_VST4_NC0_Q4_AIA_16(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)   \
_ARM_IF_ANDI_NE0(TRR,NRR,7, 11000f)                                 \
    _ARM_VST4_N_Q4_AIA_16(TRR,VRI0,VRI1,VRI2,VRI3,PRR)              \
    _ARM_GOTO_IMM(11015f)                                           \
_ARM_ENDIF(11000)                                                   \
    _ARM_VST4_Q4_AIA(16, VRI0,VRI1,VRI2,VRI3,PRR)                   \
"11015: \n\t"

# define _ARM_VST4_NC0_Q4_AIA_32(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)   \
_ARM_IF_ANDI_NE0(TRR,NRR,3, 11000f)                                 \
    _ARM_VST4_N_Q4_AIA_32(TRR,VRI0,VRI1,VRI2,VRI3,PRR)              \
    _ARM_GOTO_IMM(11015f)                                           \
_ARM_ENDIF(11000)                                                   \
    _ARM_VST4_Q4_AIA(32, VRI0,VRI1,VRI2,VRI3,PRR)                   \
"11015: \n\t"



# define _ARM_VST4_NC0B_D4_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)   \
	_ARM_VST4_NC0_D4_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)

# define _ARM_VST4_NC0B_D4_AIA_16(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,1)                                         \
	_ARM_VST4_NC0_D4_AIA_16(TRR,TRR,VRI0,VRI1,VRI2,VRI3,PRR)

# define _ARM_VST4_NC0B_D4_AIA_32(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,2)                                         \
	_ARM_VST4_NC0_D4_AIA_32(TRR,TRR,VRI0,VRI1,VRI2,VRI3,PRR)


# define _ARM_VST4_NC0B_Q4_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)   \
	_ARM_VST4_NC0_Q4_AIA_8(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)

# define _ARM_VST4_NC0B_Q4_AIA_16(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,1)                                         \
	_ARM_VST4_NC0_Q4_AIA_16(TRR,TRR,VRI0,VRI1,VRI2,VRI3,PRR)

# define _ARM_VST4_NC0B_Q4_AIA_32(TRR,NRR,VRI0,VRI1,VRI2,VRI3,PRR)  \
	_ARM_RSHIFTI(TRR,NRR,2)                                         \
	_ARM_VST4_NC0_Q4_AIA_32(TRR,TRR,VRI0,VRI1,VRI2,VRI3,PRR)

/* @} */

/* ========================================================================= */

/*
 * Store narrow (truncate) multiple N-element structures
 */

        /*
         * Store only 1st byte from data-type
         */

# define _ARM_VST1U8_Q1(W, VRI0,PRR)                    VIPM_GLUE(_ARM_VST1U8_Q1_,W)(W, VRI0,PRR)
# define _ARM_VST1U8_Q2(W, VRI0,VRI1,PRR)               VIPM_GLUE(_ARM_VST1U8_Q2_,W)(W, VRI0,VRI1,PRR)

# define _ARM_VST1U8_Q1_AIA(W, VRI0,PRR)                VIPM_GLUE(_ARM_VST1U8_Q1_AIA_,W)(W, VRI0,PRR)
# define _ARM_VST1U8_Q2_AIA(W, VRI0,VRI1,PRR)           VIPM_GLUE(_ARM_VST1U8_Q2_AIA_,W)(W, VRI0,VRI1,PRR)
# define _ARM_VST1U8_Q3_AIA(W, VRI0,VRI1,VRI2,PRR)      VIPM_GLUE(_ARM_VST1U8_Q3_AIA_,W)(W, VRI0,VRI1,VRI2,PRR)
# define _ARM_VST1U8_Q4_AIA(W, VRI0,VRI1,VRI2,VRI3,PRR) VIPM_GLUE(_ARM_VST1U8_Q4_AIA_,W)(W, VRI0,VRI1,VRI2,VRI3,PRR)

            /* 8bit --> 8bit */
# define _ARM_VST1U8_Q1_8(W, VRI0,PRR)                      _ARM_VST1_Q1(W, VRI0,PRR)
# define _ARM_VST1U8_Q2_8(W, VRI0,VRI1,PRR)                 _ARM_VST1_Q2(W, VRI0,VRI1,PRR)

# define _ARM_VST1U8_Q1_AIA_8(W, VRI0,PRR)                  _ARM_VST1_Q1_AIA(W, VRI0,PRR)
# define _ARM_VST1U8_Q2_AIA_8(W, VRI0,VRI1,PRR)             _ARM_VST1_Q2_AIA(W, VRI0,VRI1,PRR)
# define _ARM_VST1U8_Q3_AIA_8(W, VRI0,VRI1,VRI2,PRR)        _ARM_VST1_Q3_AIA(W, VRI0,VRI1,VRI2,PRR)
# define _ARM_VST1U8_Q4_AIA_8(W, VRI0,VRI1,VRI2,VRI3,PRR)   _ARM_VST1_Q4_AIA(W, VRI0,VRI1,VRI2,VRI3,PRR)

            /* 16bit[0] --> 8bit */
# define _ARM_VST1U8_Q1_16(W, VRI0,PRR)                      \
        _ARM_VNARROW_Q1(16, VRI0)                            \
        _ARM_VST1_D1(8, _ARM_QDMAP1(VRI0),PRR)
# define _ARM_VST1U8_Q2_16(W, VRI0,VRI1,PRR)                 \
        _ARM_VNARROW_Q2(16, VRI0,VRI1)                       \
        _ARM_VST1_Q1(8, VRI1,PRR)
# define _ARM_VST1U8_Q3_16(W, VRI0,VRI1,VRI2,PRR)            \
        _ARM_VNARROW_Q2(16, VRI0,VRI1)                       \
        _ARM_VNARROW_Q1(16, VRI2)                            \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI2), _ARM_QDMAP1(VRI2))    \
        _ARM_VST1_D3(8, _ARM_QDMAP0(VRI1),_ARM_QDMAP1(VRI1),_ARM_QDMAP0(VRI2),PRR)
# define _ARM_VST1U8_Q4_16(W, VRI0,VRI1,VRI2,VRI3,PRR)       \
        _ARM_VNARROW_Q2(16, VRI0,VRI1)                       \
        _ARM_VNARROW_Q2(16, VRI2,VRI3)                       \
        _ARM_VMOV_Q(VRI2, VRI1)                              \
        _ARM_VST1_Q2(8, VRI2,VRI3,PRR)

# define _ARM_VST1U8_Q1_AIA_16(W, VRI0,PRR)                  \
        _ARM_VNARROW_Q1(16, VRI0)                            \
        _ARM_VST1_D1_AIA(8, _ARM_QDMAP1(VRI0),PRR)
# define _ARM_VST1U8_Q2_AIA_16(W, VRI0,VRI1,PRR)             \
        _ARM_VNARROW_Q2(16, VRI0,VRI1)                       \
        _ARM_VST1_Q1_AIA(8, VRI1,PRR)
# define _ARM_VST1U8_Q3_AIA_16(W, VRI0,VRI1,VRI2,PRR)        \
        _ARM_VNARROW_Q2(16, VRI0,VRI1)                       \
        _ARM_VNARROW_Q1(16, VRI2)                            \
        _ARM_VMOV_D(_ARM_QDMAP0(VRI2), _ARM_QDMAP1(VRI2))    \
        _ARM_VST1_D3_AIA(8, _ARM_QDMAP0(VRI1),_ARM_QDMAP1(VRI1),_ARM_QDMAP0(VRI2),PRR)
# define _ARM_VST1U8_Q4_AIA_16(W, VRI0,VRI1,VRI2,VRI3,PRR)   \
        _ARM_VNARROW_Q2(16, VRI0,VRI1)                       \
        _ARM_VNARROW_Q2(16, VRI2,VRI3)                       \
        _ARM_VMOV_Q(VRI2, VRI1)                              \
        _ARM_VST1_Q2_AIA(8, VRI2,VRI3,PRR)

            /* 32bit[0] --> 8bit */
# define _ARM_VST1U8_Q1_32(W, VRI0,PRR)                      \
        _ARM_VST1_LANE_D1(8, 0,_ARM_QDMAP0(VRI0),PRR)        \
        _ARM_VST1_LANE_D1(8, 4,_ARM_QDMAP0(VRI0),PRR)        \
        _ARM_VST1_LANE_D1(8, 0,_ARM_QDMAP1(VRI0),PRR)        \
        _ARM_VST1_LANE_D1(8, 4,_ARM_QDMAP1(VRI0),PRR)
# define _ARM_VST1U8_Q2_32(W, VRI0,VRI1,PRR)                 \
        _ARM_VNARROW_Q2(32, VRI0,VRI1)                       \
        _ARM_VST1U8_Q1_16(16, VRI1,PRR)
# define _ARM_VST1U8_Q4_32(W, VRI0,VRI1,VRI2,VRI3,PRR)       \
        _ARM_VNARROW_Q2(32, VRI0,VRI1)                       \
        _ARM_VNARROW_Q2(32, VRI2,VRI3)                       \
        _ARM_VST1U8_Q2_16(W, VRI1,VRI3,PRR)

# define _ARM_VST1U8_Q1_AIA_32(W, VRI0,PRR)                  \
        _ARM_VST1_LANE_D1_AIA(8, 0,_ARM_QDMAP0(VRI0),PRR)    \
        _ARM_VST1_LANE_D1_AIA(8, 4,_ARM_QDMAP0(VRI0),PRR)    \
        _ARM_VST1_LANE_D1_AIA(8, 0,_ARM_QDMAP1(VRI0),PRR)    \
        _ARM_VST1_LANE_D1_AIA(8, 4,_ARM_QDMAP1(VRI0),PRR)
# define _ARM_VST1U8_Q2_AIA_32(W, VRI0,VRI1,PRR)             \
        _ARM_VNARROW_Q2(32, VRI0,VRI1)                       \
        _ARM_VST1U8_Q1_AIA_16(16, VRI1,PRR)
# define _ARM_VST1U8_Q3_AIA_32(W, VRI0,VRI1,VRI2,PRR)        \
        _ARM_VST1U8_Q2_AIA_32(W, VRI0,VRI1,PRR)              \
        _ARM_VST1U8_Q1_AIA_32(W, VRI2,PRR)
# define _ARM_VST1U8_Q4_AIA_32(W, VRI0,VRI1,VRI2,VRI3,PRR)   \
        _ARM_VNARROW_Q2(32, VRI0,VRI1)                       \
        _ARM_VNARROW_Q2(32, VRI2,VRI3)                       \
        _ARM_VST1U8_Q2_AIA_16(W, VRI1,VRI3,PRR)

/* @} */


/* ========================================================================= */

/*
 * Print
 * @{
 */

# define _ARM_VPRINT_LANE_D1(W,_Fmt,L,VRI) \
    _ARM_VPUSH_ALL()                       \
    _ARM_PUSH_ALLRR()                      \
	__ARM_VPRINT_LANE_D1(W,_Fmt,L,VRI)     \
	_ARM_POP_ALLRR()                       \
	_ARM_VPOP_ALL()

# define _ARM_VPRINTF32_LANE_D1(_Fmt,L,VRI) \
    _ARM_VPUSH_ALL()                        \
    _ARM_PUSH_ALLRR()                       \
	__ARM_VPRINTF32_LANE_D1(_Fmt,L,VRI)	    \
	_ARM_POP_ALLRR()                        \
	_ARM_VPOP_ALL()

/* @} */
/* @} */


/* ######################################################################### */

# include "Aas.h"

#endif /* !defined(__VIPM_ARM_ASM__) */


/******************************************************************************
 *
 *
 *
 *****************************************************************************/
