/* -*- Mode: c; tab-width: 4; -*-
 * -*- ex: ts=4 -*-
 */

/*
 * In this example is shown static recognition mode
 */

/******************************************************************************
 * %% BeginSection: includes
 */

/* for libBo and Windows compatibility */
#define HAS_TIMEVAL_STRUC

#include <Bo/libc/Stdio.h>
#include <Bo/libc/Limits.h>
	/* BoL_MB_LEN_MAX */
#include <Bo/aorp/Mld.h>
#define __NEED_UNIX_ERRORS
#include <Bo/Errors.h>
#include <Bo/services/Ccsconv.h>
#include <Bo/services/Err.h>

#include <Vodi/Vodilib.h>
#include <Vodi/objects/Vpwprinc.h>
#include <Vodi/services/Ucntl.h>
#include <Vodi/services/Vodiprinc.h>
#include <Vodi/services/Vodiens.h>
#include <Vodi/services/Vodires.h>
#include <Vodi/utils/arrio/Filepath.h>

/******************************************************************************
 * %% BeginSection: local function prototypes
 */
#pragma mark -
#pragma mark [ local function prototypes ]
#pragma mark -

static bo_status_t _t_load_requireds(struct aorp_error *);

struct result_handler {
	struct aorp_foreach_pred _h;
	struct aorp_error        *err;
};

static bo_status_t
	_t_result_handle(aorp_object_t, struct result_handler *),
	_t_plateinfo_handle(struct vodi_vpw_result_info *, struct aorp_error *);

/******************************************************************************
 * %% BeginSection: main
 */
#pragma mark -
#pragma mark [ main ]
#pragma mark -

int main(int argc, char *argv[])
{
	bo_status_t status;
	aorp_object_t principal, ensemble;
	char const *path;
	struct vodi_image *image;

	struct aorp_error *err;
	struct aorp_error256 err256;

	if (1 == argc) {
		fprintf(stderr,
			"Please specify one or more images\n"
			"For example: %s pic1.bmp pic2.jpg pic3.jpg\n",
			argv[0]
		);
		return (0);
		/* NOTREACHED */
	}

	_AORP_ERROR256_Init(&err256);
	err = (struct aorp_error *)&err256;

	principal = NULL;

	status = _t_load_requireds(err);
	if (BoS_FAILURE(status)) {
		goto L_error;
		/* NOTREACHED */
	}

	principal = VpwprincOpen(0, NULL, NULL, err);
	if (NULL == principal) {
		goto L_error;
		/* NOTREACHED */
	}

	++argv; /* skip self */
	while (NULL != (path = *argv++)) {
		if (NULL == (image = VodiARRIOPgetimage(path, NULL, err))) {
			status = BoS_ERR;
			goto L_error;
			/* NOTREACHED */
		}
		status = VodiprincProcess(
			principal, image, 0, NULL, NULL, NULL, &ensemble, NULL, err
		);
		(void)VodiImageRelease(image, 0, NULL);
		if (BoS_FAILURE(status)) {
			goto L_error;
			/* NOTREACHED */
		}
		if (NULL != ensemble) {
			struct result_handler handler = {
				{ AorpK_HINT_FOREACHPRED, 0,
				(aorp_feachpred_cb)_t_result_handle }, err
			};
			status = VodiensForeach(ensemble, &handler, err);
			AorpRls(ensemble);
			if (BoS_FAILURE(status)) {
				goto L_error;
				/* NOTREACHED */
			}
		}
	}

 L_finally:
	AorpRls(principal);

	return (0);
	/* NOTREACHED */

 L_error:
	fprintf(stderr, "%s\n", err->msg);
	goto L_finally;
	/* NOTREACHED */
}

/******************************************************************************
 * %% BeginSection: local function definitions
 */
#pragma mark -
#pragma mark [ local function definitions ]
#pragma mark -

static bo_status_t _t_load_requireds(struct aorp_error *anErrPtr)
{
	/*
	 * Loading of 2 necessary modules:
	 * vpw  - basic functional
	 * vpwi - description of countries (load only needed set)
	 */
	static char const *_s_requireds[] = {
		"vpw", "vpwi-ru", NULL
	};

	bo_status_t status;
	char const *name, **rqp;

	for (rqp = _s_requireds; NULL != (name = *rqp); ++rqp) {
		status = AorpMldLoad(name, NULL, 0, 0, anErrPtr);
		if (BoS_FAILURE(status)) {
			return (status);
			/* NOTREACHED */
		}
	}
	return (BoS_NORMAL);
}

static bo_status_t
_t_result_handle(aorp_object_t aResult, struct result_handler *aHandler)
{
	bo_status_t status;
	struct vodi_result_info *info;
	struct vodi_vpw_result_info plateinfo;

	info = (struct vodi_result_info *)&plateinfo;
	VODI_RESULT_INFO_TYPE(info) = VodiK_VPW_RESULT_INFO;
	status = VodiresFetchinfo(aResult, info, aHandler->err);
	if (BoS_SUCCESS(status)) {
		status = _t_plateinfo_handle(&plateinfo, aHandler->err);
		(void)VodiResultInfoDestroy(info);
	}
	return (BoS_SUCCESS(status) ? BoS_OK : status);
}

static bo_status_t
_t_plateinfo_handle(
	struct vodi_vpw_result_info *anInfo,
	struct aorp_error *anErrPtr
	)
{
	bo_status_t status;
	struct vodi_plate_info_spec *pis;
	struct vodi_plate *plate;
	char plateStr[VodiL_VPW_SYMBOL_MAX * BoL_MB_LEN_MAX];
	struct aorp_ccscvtor cconv;
	bo_cset_t outpenc;

	pis = VODI_RESULT_INFO_SPEC(anInfo, plate);
	plate = &pis->pis_plate_variantv[0];
	/* BoK_UTF8_CSET, BoK_IBM_CYR_CSET, BoK_WIN_CYR_CSET, BoK_KOI8U_CSET */
#ifndef BO_SYSTYPE_WIN32
	outpenc = BoK_UTF8_CSET;
#else
	outpenc = BoK_IBM_CYR_CSET;
#endif
	status = AorpOpenCcscvtor_i_(BoK_UNK_CSET, outpenc, &cconv, anErrPtr);
	if (BoS_SUCCESS(status)) {
		(void)AorpCcsconv_s_2mb(&cconv, plate->pv_plate_string, plateStr, NULL);
		printf("%s\n", plateStr);
		(void)AorpCcscvtorClose(&cconv, NULL);
	}
	return (status);
}

/******************************************************************************
 *
 *
 *****************************************************************************/
