/* -*- Mode: c; tab-width: 4; -*-
 * -*- ex: ts=4 -*-
 */

/******************************************************************************
 * %% BeginSection: includes
 */

/* for libBo and Windows compatibility */
#define HAS_TIMEVAL_STRUC

#include <Bo/libc/Strings.h>
#include <Bo/libc/Stdio.h>
#include <Bo/libc/Limits.h>
	/* BoL_MB_LEN_MAX */
#include <Bo/libc/Fmtio.h>
#include <Bo/aorp/Mld.h>
#define __NEED_UNIX_ERRORS
#include <Bo/Errors.h>
#include <Bo/services/Ccsconv.h>
#include <Bo/services/Err.h>
#include <Bo/memman/BoMalloc.h>
#include <Bo/memman/Zmalloc.h>
#include <Bo/fundis/Seqlist.h>

#include <Vodi/Vodilib.h>
#include <Vodi/objects/Vpwprinc.h>
#include <Vodi/services/Ucntl.h>
#include <Vodi/services/Vodiprinc.h>
#include <Vodi/services/Vodiens.h>
#include <Vodi/services/Vodires.h>
#include <Vodi/utils/arrio/Filepath.h>

/******************************************************************************
 * %% BeginSection: type/constant declarations
 */
#pragma mark -
#pragma mark [ type/constant declarations ]
#pragma mark -

/* See also <Bo/sys/csType.h> */
#ifndef BO_SYSTYPE_WIN32
# define _K_OUTPUT_CSET BoK_UTF8_CSET
#else
# define _K_OUTPUT_CSET BoK_IBM_CYR_CSET
#endif

#define _K_TRACK_SIZE_MAX 512

struct track_elem {
	int        x;
	int        y;
	unsigned   i; /* frame sequence number */
	bo_utime_t t; /* frame timestamp */
};

typedef bo_seqlist_(struct track_elem) track_t;
#define _t_track_size(x)         _BO_SEQLIST_Size(x)
#define _t_track_begin(x)        BO_SEQLIST_Begin(struct track_elem,x)
#define _t_track_end(x)          BO_SEQLIST_End(struct track_elem,x)
#define _t_track_push(x,y,z)     BO_SEQLIST_PushX2(struct track_elem,x,y,z)
#define _t_track_append(x,y,z)   BO_SEQLIST_Append_s(struct track_elem,x,y,z)
#define _t_track_assign(x,y,z)   BO_SEQLIST_Assign_s(struct track_elem,x,y,z)
#define _t_track_popfront_n(x,y) BO_SEQLIST_PopFront_n(struct track_elem,x,y)

#ifndef BO_SEQLIST_PopFront_n
# define BO_SEQLIST_PopFront_n(_TListValueType,aList,N) \
	do { \
		struct bo_seqlist *__popfront_list = bo_seqlist_p(aList); \
		size_t __popfront_nelems = (N); \
		size_t __popfront_nbytes = sizeof(_TListValueType) * __popfront_nelems; \
		(void)memmove(__popfront_list->memory, \
			(char *)__popfront_list->memory + __popfront_nbytes, \
			__popfront_nbytes); \
		__popfront_list->elemc -= __popfront_nelems; \
	} while (0)
#endif

/******************************************************************************
 * %% BeginSection: local function prototypes
 */
#pragma mark -
#pragma mark [ local function prototypes ]
#pragma mark -

static bo_status_t _t_load_modules(struct aorp_error *);
static void _t_unload_modules();

static bo_status_t _t_princ_configure(aorp_object_t, struct aorp_error *);

struct result_handler {
	struct aorp_foreach_pred _h;
	struct aorp_error        *err;
};

static bo_status_t
	_t_ensemble_handle(aorp_object_t, struct aorp_error *),
	_t_result_handle(aorp_object_t, struct result_handler *),
	_t_plateinfo_handle(struct vodi_vpw_result_info *, struct aorp_error *);

static bo_status_t
	_t_dyn_rinit(struct vodi_result_info *, bo_pointer_t, struct aorp_error *),
	_t_dyn_rfold(struct vodi_result_info *, struct vodi_result_info *,
		bo_pointer_t, struct aorp_error *);

static struct bo_zmalloc *
_t_track_alloc_open(struct aorp_error *);

static bo_status_t
	_t_track_init(bo_pointer_t, track_t *),
	_t_track_fini(bo_pointer_t, track_t *);

static bo_status_t
_t_vodi_ucontext_init(vodi_ucontext_t *, struct bo_zmalloc *, struct aorp_error *);

static ssize_t
	_t_vodi_ucontext_retain(track_t *, struct aorp_error *),
	_t_vodi_ucontext_release(track_t *, bo_opflags_t, struct aorp_error *);

static bo_status_t
_t_vodi_ucontext_dup(vodi_ucontext_t const *, vodi_ucontext_t *, struct aorp_error *);

/******************************************************************************
 * %% BeginSection: main
 */
#pragma mark -
#pragma mark [ main ]
#pragma mark -

int main(int argc, char *argv[])
{
	bo_status_t status;
	int argi;
	aorp_object_t principal, ensemble;
	struct vodi_image *image;
	struct vodi_vpw_options options = {0};
	struct bo_zmalloc *trackAllocator;
	vodi_ucontext_t uctx;

	struct aorp_error *err;
	struct aorp_error256 err256;

	if (1 == argc) {
		fprintf(stderr,
			"Please specify one or more images\n"
			"For example: %s pic1.jpg pic2.jpg pic3.jpg\n",
			argv[0]
		);
		return (0);
		/* NOTREACHED */
	}

	_AORP_ERROR256_Init(&err256);
	err = (struct aorp_error *)&err256;

	_G_BoMalloc_verbose = 1;

	principal = NULL;
	trackAllocator = NULL;

	status = _t_load_modules(err);
	if (BoS_FAILURE(status)) {
		goto L_error;
		/* NOTREACHED */
	}

	trackAllocator = _t_track_alloc_open(err);
	status = (NULL != trackAllocator ? BoS_NORMAL : BoS_ERR);
	if (BoS_FAILURE(status)) {
		goto L_error;
		/* NOTREACHED */
	}

	principal = VpwprincOpen(0, NULL, NULL, err);
	status = (NULL != principal ? BoS_NORMAL : BoS_ERR);
	if (BoS_FAILURE(status)) {
		goto L_error;
		/* NOTREACHED */
	}

	status = _t_princ_configure(principal, err);
	if (BoS_FAILURE(status)) {
		goto L_error;
		/* NOTREACHED */
	}

	for (argi = 1; argi < argc; ++argi) {
		image = VodiARRIOPgetimage(argv[argi], NULL, err);
		status = (NULL != image ? BoS_NORMAL : BoS_ERR);
		if (BoS_FAILURE(status)) {
			goto L_error;
			/* NOTREACHED */
		}
		status = _t_vodi_ucontext_init(&uctx, trackAllocator, err);
		if (BoS_FAILURE(status)) {
			(void)VodiImageRelease(image, 0, NULL);
			goto L_error;
			/* NOTREACHED */
		}
		options.img_seqnum++;
		options.img_timestamp += 40000; /* 40000 usec = 40 msec */
		status = VodiprincProcess(
			principal, image, 0, NULL, &uctx, NULL, &ensemble, &options, err
		);
		VODI_UCONTEXT_DRelease(&uctx, 0, NULL);
		(void)VodiImageRelease(image, 0, NULL);
		if (BoS_FAILURE(status)) {
			goto L_error;
			/* NOTREACHED */
		}
		status = _t_ensemble_handle(ensemble, err);
		AorpRls(ensemble);
		if (BoS_FAILURE(status)) {
			goto L_error;
			/* NOTREACHED */
		}
	}

	/* in dynamic mode it's necessary to flush principal */
	status = VodiprincFlush(principal, NULL, &ensemble, err);
	if (BoS_FAILURE(status)) {
		goto L_error;
		/* NOTREACHED */
	}
	status = _t_ensemble_handle(ensemble, err);
	AorpRls(ensemble);
	if (BoS_FAILURE(status)) {
		goto L_error;
		/* NOTREACHED */
	}

 L_finally:
	AorpRls(principal);
	(void)_BoZmallocClose(trackAllocator, 0, NULL);
	_t_unload_modules();

	return (0);
	/* NOTREACHED */

 L_error:
	fprintf(stderr, "%s\n", err->msg);
	goto L_finally;
	/* NOTREACHED */
}

/******************************************************************************
 * %% BeginSection: local function definitions
 */
#pragma mark -
#pragma mark [ local function definitions ]
#pragma mark -

/*
 * Loading of 2 necessary modules:
 * vpw  - basic functional
 * vpwi - description of countries (load only needed set)
 */
static char const *_S_modules[] = {
	"vpw", "vpwi-ru", NULL
};

static bo_status_t _t_load_modules(struct aorp_error *anErrPtr)
{
	bo_status_t status;
	char const **mp;
	status = BoS_NORMAL;
	for (mp = _S_modules; BoS_SUCCESS(status) && NULL != *mp; ++mp)
		status = AorpMldLoad(*mp, NULL, 0, 0, anErrPtr);
	return (status);
}

static void _t_unload_modules()
{
	char const **mp;
	for (mp = _S_modules; NULL != *mp; ++mp);
	for (--mp; _S_modules <= mp; --mp)
		(void)AorpMldUnload(*mp, NULL);
}

/* -------------------------------------------------------------------------- */

static bo_status_t
_t_princ_configure(aorp_object_t aPrincipal, struct aorp_error *anErrPtr)
{
	bo_status_t status;

	status = BoS_NORMAL;
	status |= VodiprincSetparam(aPrincipal, anErrPtr,
		VodiCTL_VPW_DYNAMIC_ENABLE, 1);
	status |= VodiprincSetparam(aPrincipal, anErrPtr,
		VodiCTL_VPW_DYNAMIC_RINIT_CALLBACK, _t_dyn_rinit);
	status |= VodiprincSetparam(aPrincipal, anErrPtr,
		VodiCTL_VPW_DYNAMIC_SUPERPOSITION_CALLBACK, _t_dyn_rfold);

	/* other parameters */

	status |= VodiprincApplyparam(aPrincipal, anErrPtr);
	return (status);
}

/* -------------------------------------------------------------------------- */

static bo_status_t
_t_ensemble_handle(aorp_object_t anEnsemble, struct aorp_error *anErrPtr)
{
	bo_status_t status;
	struct result_handler handler = {
		{ AorpK_HINT_FOREACHPRED, 0,
		(aorp_feachpred_cb)_t_result_handle }, anErrPtr
	};

	status = BoS_NORMAL;
	if (NULL != anEnsemble)
		status = VodiensForeach(anEnsemble, &handler, anErrPtr);
	return (status);
}

static bo_status_t
_t_result_handle(aorp_object_t aResult, struct result_handler *aHandler)
{
	bo_status_t status;
	struct vodi_result_info *info;
	struct vodi_vpw_result_info plateinfo;

	info = (struct vodi_result_info *)&plateinfo;
	VODI_RESULT_INFO_TYPE(info) = VodiK_VPW_RESULT_INFO;
	status = VodiresFetchinfo(aResult, info, aHandler->err);
	if (BoS_SUCCESS(status)) {
		status = _t_plateinfo_handle(&plateinfo, aHandler->err);
		(void)VodiResultInfoDestroy(info);
	}
	return (BoS_SUCCESS(status) ? BoS_OK : status);
}

static bo_status_t
_t_plateinfo_handle(
	struct vodi_vpw_result_info *anInfo,
	struct aorp_error *anErrPtr
	)
{
	bo_status_t status;
	struct vodi_plate_info_spec *pis;
	struct vodi_plate *plate;
	char plateStr[VodiL_VPW_SYMBOL_MAX * BoL_MB_LEN_MAX];
	struct aorp_ccscvtor cconv;
	track_t *track;
	struct track_elem *te;

	pis = VODI_RESULT_INFO_SPEC(anInfo, plate);
	track = pis->pis_uctx.uctx_udata;
	plate = &pis->pis_plate_variantv[0];

	if (!(VodiF_RESULT_LOST & pis->pis_flags)) {
		return (BoS_NORMAL);
		/* NOTREACHED */
	}

	status = AorpOpenCcscvtor_i_(BoK_UNK_CSET, _K_OUTPUT_CSET, &cconv, anErrPtr);
	if (BoS_SUCCESS(status)) {
		(void)AorpCcsconv_s_2mb(&cconv, plate->pv_plate_string, plateStr, NULL);
		printf("%s:[", plateStr);
		for (te = _t_track_begin(track); te != _t_track_end(track); ++te) {
			printf("{x:%d,y:%d,i:%u,t:%" PRIbtime "}%s",
				te->x, te->y, te->i, te->t,
				te + 1 != _t_track_end(track) ? "," : "");
		}
		printf("]\n");
		(void)AorpCcscvtorClose(&cconv, NULL);
	}
	return (status);
}

/* -------------------------------------------------------------------------- */

static bo_status_t
_t_dyn_rinit(
	struct vodi_result_info *aResultInfo,
	bo_pointer_t anUserCtx,
	struct aorp_error *anErrPtr
	)
{
	struct vodi_plate_info_spec *pis;
	track_t *track;
	vodi_plate_t *plate;
	vodi_rect_t *rect;
	struct track_elem elem;

	pis = VODI_RESULT_INFO_SPEC(aResultInfo, plate);
	plate = &pis->pis_plate_variantv[0];
	rect = &plate->pv_exact_rect;
	track = pis->pis_uctx.uctx_udata;

	elem.x = (rect->rc_left + rect->rc_right) / 2;
	elem.y = (rect->rc_top + rect->rc_bottom) / 2;
	elem.i = pis->pis_seqnum;
	elem.t = pis->pis_ctime;
	_t_track_push(track, elem, { return BoS_ERR; });
	return (BoS_NORMAL);
}

static bo_status_t
_t_dyn_rfold(
	struct vodi_result_info *aSrcDst,
	struct vodi_result_info *aSrc,
	bo_pointer_t anUserCtx,
	struct aorp_error *anErrPtr
	)
{
	struct vodi_plate_info_spec *pisSrcDst, *pisSrc;
	track_t *trackSrcDst, *trackSrc;

	pisSrcDst = VODI_RESULT_INFO_SPEC(aSrcDst, plate);
	trackSrcDst = pisSrcDst->pis_uctx.uctx_udata;
	pisSrc = VODI_RESULT_INFO_SPEC(aSrc, plate);
	trackSrc = pisSrc->pis_uctx.uctx_udata;

	_t_track_append(trackSrcDst, trackSrc, { return BoS_ERR; });
	if (_K_TRACK_SIZE_MAX < _t_track_size(trackSrcDst))
		_t_track_popfront_n(trackSrcDst, _t_track_size(trackSrcDst) - _K_TRACK_SIZE_MAX);

	return (BoS_NORMAL);
}

/* -------------------------------------------------------------------------- */

static struct bo_zmalloc *_t_track_alloc_open(struct aorp_error *anErrPtr)
{
	struct bo_zmalloc_uctx ctx = {0};
	ctx.zmu_uinit_once = NULLF;
	ctx.zmu_ufini_once = NULLF;
	ctx.zmu_uinit = (bo_zmalloc_uinit_fn)_t_track_init;
	ctx.zmu_ufini = (bo_zmalloc_ufini_fn)_t_track_fini;
	ctx.zmu_ext = NULL;
	return _BoZmallocOpen(NULL, NULL, &ctx, 0, sizeof(track_t), anErrPtr);
}

static bo_status_t _t_track_init(bo_pointer_t _, track_t *aTrack)
{
	_BO_SEQLIST_Init(aTrack);
	return (BoS_NORMAL);
}

static bo_status_t _t_track_fini(bo_pointer_t _, track_t *aTrack)
{
	_BO_SEQLIST_Destroy_s(aTrack);
	return (BoS_NORMAL);
}

/* -------------------------------------------------------------------------- */

static bo_status_t
_t_vodi_ucontext_init(
	vodi_ucontext_t *aCtx,
	struct bo_zmalloc *aTrackAllocator,
	struct aorp_error *anErrPtr
	)
{
	aCtx->uctx_dup = (vodi_uctx_dup_fn)_t_vodi_ucontext_dup;
	aCtx->uctx_dretain = (vodi_udata_retain_fn)_t_vodi_ucontext_retain;
	aCtx->uctx_drelease = (vodi_udata_release_fn)_t_vodi_ucontext_release;
	aCtx->uctx_udata = _BoZmallocGet(aTrackAllocator, anErrPtr);
	return (NULL != aCtx->uctx_udata ? BoS_NORMAL : BoS_ERR);
}

static ssize_t
_t_vodi_ucontext_retain(track_t *aTrack, struct aorp_error *_)
{
	return (ssize_t)_BoZmPtrRetain(aTrack);
}

static ssize_t
_t_vodi_ucontext_release(track_t *aTrack, bo_opflags_t _, struct aorp_error *__)
{
	return (ssize_t)_BoZmPtrRelease(aTrack);
}

static bo_status_t
_t_vodi_ucontext_dup(
	vodi_ucontext_t const *aSrc,
	vodi_ucontext_t *aDst,
	struct aorp_error *anErrPtr
	)
{
	*aDst = *aSrc;
	aDst->uctx_udata = _BoZmallocGet(_BoZmPtrOwner(aSrc->uctx_udata), anErrPtr);
	if (NULL == aDst->uctx_udata) {
		return (BoS_ERR);
		/* NOTREACHED */
	}
	_t_track_assign(aDst->uctx_udata, aSrc->uctx_udata,
		{
			(void)_BoZmPtrRelease(aDst->uctx_udata);
			return (BoS_ERR);
			/* NOTREACHED */
		}
	);
	return (BoS_NORMAL);
}

/******************************************************************************
 *
 *
 *****************************************************************************/
